# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


from struct import pack, unpack

TAG_ZERO = 0
TAG_START_OBJ = 1
TAG_END_OBJ = 2
TAG_CHAR = 3
TAG_UNSIGNED_CHAR = 4
TAG_INT = 5
TAG_UNSIGNED_INT = 6
TAG_SHORT = 7
TAG_UNSIGNED_SHORT = 8
TAG_LONG = 9
TAG_UNSIGNED_LONG = 10
TAG_LONG_LONG = 11
TAG_UNSIGNED_LONG_LONG = 12
TAG_FLOAT = 13
TAG_DOUBLE = 14
TAG_STRING = 15
TAG_BLOB = 16
TAG_EXCEPTION = 17
TAG_START_REC = 18
TAG_END_REC = 19
TAG_EOF = 20
TAG_LARGE_BLOB = 21  # For blobs >= 2Gb
LAST_TAG = 22

TAG_NAME = (
    "0",
    "start of object",
    "end of object",
    "char",
    "unsigned char",
    "int",
    "unsigned int",
    "short",
    "unsigned short",
    "long",
    "unsigned long",
    "long long",
    "unsigned long long",
    "float",
    "double",
    "string",
    "blob",
    "exception",
    "start of record",
    "end of record",
)


class Stream:
    def __init__(self, stream):
        self.stream = stream

    def _read(self, n):
        return self.stream.read(n)

    def read_tag(self, expected_tag):
        tag = self._read(1)
        tag = ord(tag)

        while tag == TAG_END_OBJ:
            tag = self._read(1)
            if tag:
                tag = ord(tag)

        # TODO There is some suspicious code on the perl side for the following.
        # TODO Maybe worth having a look at it sometime.
        if tag != expected_tag:
            if tag < len(TAG_NAME):
                raise RuntimeError(
                    "Unexpected tag: %s, wanted %s"
                    % (TAG_NAME[tag], TAG_NAME[expected_tag])
                )
            else:
                raise RuntimeError("Invalid tag with id: '%s'" % tag)

        return tag

    def write_tag(self, tag):
        self._write(pack("b", tag), 1)

    def next_object(self):
        while 1:
            tag = self.stream.recv(1)
            if tag == "":
                return 0

            tag = unpack("b", tag)[0]

            if tag == TAG_START_OBJ:
                return 1
            else:
                if tag < len(TAG_NAME):
                    raise RuntimeError(
                        "Unexpected tag: '%s', wanted '%s'"
                        % (TAG_NAME[tag], TAG_NAME[TAG_START_OBJ])
                    )
                else:
                    raise RuntimeError("Invalid tag with id: '%d'" % tag)

    def write_char(self, c):
        self.write_tag(TAG_CHAR)
        self._write(c, 1)

    def write_unsigned_char(self, c):
        self.write_tag(TAG_UNSIGNED_CHAR)
        self._write(c, 1)

    def write_int(self, n):
        self.write_tag(TAG_INT)
        self._write(pack("!L", n), 4)

    def write_unsigned_int(self, n):
        self.write_tag(TAG_UNSIGNED_INT)
        self._write(pack("!L", n), 4)

    def write_long(self, n):
        self.write_tag(TAG_LONG)
        self._write(pack("!L", n), 4)

    def write_double(self, n):
        self.write_tag(TAG_DOUBLE)
        raise RuntimeError("write_doble")

    def write_unsigned_long(self, n):
        self.write_tag(TAG_UNSIGNED_LONG)
        self._write(pack("!L", n), 4)

    def write_long_long(self, n):
        self.write_tag(TAG_LONG_LONG)

        # TODO Error handling when the number is too long.
        # TODO See the perl.

        self._write(pack("!L", n), 4)

    def write_unsigned_long_long(self, n):
        self.write_tag(TAG_UNSIGNED_LONG_LONG)

        # TODO Error handling when input value too big.

        self._write(pack("!L", 0), 4)
        self._write(pack("!L", n), 4)

    def write_string(self, s):
        self.write_tag(TAG_STRING)
        n = len(s)
        self._write(pack("!L", n), 4)
        self._write(s, n)

    def read_char(self):
        self.read_tag(TAG_CHAR)
        return self._read(1)

    def read_unsigned_char(self):
        self.read_tag(TAG_UNSIGNED_CHAR)
        return self._read(1)

    def read_int(self):
        self.read_tag(TAG_INT)
        return unpack("!L", self._read(4))[0]

    def read_unsigned_int(self):
        self.read_tag(TAG_UNSIGNED_INT)
        return unpack("!L", self._read(4))[0]

    def read_long(self):
        self.read_tag(TAG_LONG)
        return unpack("!L", self._read(4))[0]

    def read_double(self):
        self.read_tag(TAG_DOUBLE)
        raise RuntimeError("read_double")

    def read_unsigned_long(self):
        self.read_tag(TAG_UNSIGNED_LONG)
        return unpack("!L", self._read(4))[0]

    def read_long_long(self):
        self.read_tag(TAG_LONG_LONG)
        raise RuntimeError("read_long_long")

    def _read_long_long(self):
        return unpack("!Q", self._read(8))[0]
        # TODO: use 'Q' instead of 'L' for 64-bit ints
        x = unpack("!L", self._read(4))[0]
        y = unpack("!L", self._read(4))[0]

        return (x << 32) | y

    def read_unsigned_long_long(self):
        self.read_tag(TAG_UNSIGNED_LONG_LONG)
        return unpack("!Q", self._read(8))[0]

    def read_string(self):
        self.read_tag(TAG_STRING)
        n = unpack("!L", self._read(4))[0]
        return self._read(n)

    def read_large_blob(self):
        self.read_tag(TAG_LARGE_BLOB)
        len = unpack("!Q", self._read(8))[0]
        return self._read(len)

    def start_object(self):
        self.write_tag(TAG_START_OBJ)

    def end_object(self):
        self.write_tag(TAG_END_OBJ)

    def read_object(self):
        if not self.next_object():
            return None
        klass = self.read_string()
        return klass

    def write_object(self, name, ref):
        self.start_object()

        self.write_string(name)
        self.write_string(ref)

        self.end_object()



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import logging

from multiurl import Downloader

from earthkit.regrid.utils import progress_bar
from earthkit.regrid.utils.caching import SETTINGS, cache_file

LOG = logging.getLogger(__name__)


def download_and_cache(
    url,
    *,
    owner="url",
    parts=None,
    verify=True,
    force=None,
    chunk_size=1024 * 1024,
    range_method="auto",
    http_headers=None,
    update_if_out_of_date=False,
    fake_headers=None,  # When HEAD is not allowed but you know the size
    progress_bar=progress_bar,
    maximum_retries=0,
    retry_after=10,
    **kwargs,
):
    # TODO: re-enable this feature
    extension = None

    LOG.debug("URL %s", url)

    downloader = Downloader(
        url,
        chunk_size=chunk_size,
        timeout=SETTINGS.get("url-download-timeout"),
        verify=verify,
        parts=parts,
        range_method=range_method,
        http_headers=http_headers,
        fake_headers=fake_headers,
        # statistics_gatherer=record_statistics,
        # progress_bar=progress_bar,
        progress_bar=progress_bar,
        resume_transfers=True,
        override_target_file=False,
        download_file_extension=".download",
        maximum_retries=maximum_retries,
        retry_after=retry_after,
    )

    if extension and extension[0] != ".":
        extension = "." + extension

    if extension is None:
        extension = downloader.extension()

    path = downloader.local_path()
    if path is not None:
        return

    def out_of_date(url, path, cache_data):
        if SETTINGS.get("check-out-of-date-urls") is False:
            return False

        if downloader.out_of_date(path, cache_data):
            if SETTINGS.get("download-out-of-date-urls") or update_if_out_of_date:
                LOG.warning(
                    "Invalidating cache version and re-downloading %s",
                    url,
                )
                return True
            else:
                LOG.warning(
                    "To enable automatic downloading of updated URLs set the 'download-out-of-date-urls'"
                    " setting to True",
                )
        return False

    if force is None:
        force = out_of_date

    def download(target, _):
        downloader.download(target)
        return downloader.cache_data()

    path = cache_file(
        owner,
        download,
        dict(url=url, parts=parts),
        extension=extension,
        force=force,
    )

    return path



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import numpy as np
from scipy.sparse import csr_array, save_npz

from .stream import Stream


def dtype_uint(little_endian, size):
    order = "<" if little_endian else ">"
    return np.dtype({4: np.uint32}[size]).newbyteorder(order)


def dtype_float(little_endian, size):
    order = "<" if little_endian else ">"
    return np.dtype({4: np.float32, 8: np.float64}[size]).newbyteorder(order)


def mir_cached_matrix_to_array(path):
    with open(path, "rb") as f:
        s = Stream(f)
        rows = s.read_unsigned_long()  # rows
        cols = s.read_unsigned_long()  # cols
        s.read_unsigned_long()  # non-zeros, ignored

        little_endian = s.read_int() != 0  # little_endian
        index_item_size = s.read_unsigned_long()  # sizeof(index)
        scalar_item_size = s.read_unsigned_long()  # sizeof(scalar)
        s.read_unsigned_long()  # sizeof(size), ignored

        outer = s.read_large_blob()  # outer
        inner = s.read_large_blob()  # inner
        data = s.read_large_blob()  # data

        outer = np.frombuffer(
            outer,
            dtype=dtype_uint(little_endian, index_item_size),
        )

        inner = np.frombuffer(
            inner,
            dtype=dtype_uint(little_endian, index_item_size),
        )

        data = np.frombuffer(
            data,
            dtype=dtype_float(little_endian, scalar_item_size),
        )

        return csr_array((data, inner, outer), shape=(rows, cols))


def mir_cached_matrix_to_file(path, target):
    if not target.endswith(".npz"):
        raise ValueError("target must end with .npz")

    z = mir_cached_matrix_to_array(path)
    save_npz(target, z)



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


def pytest_runtest_setup(item):
    # print(f"config {item.config.option}")

    marks_in_items = list([m.name for m in item.iter_markers()])

    from earthkit.regrid.db import SYS_DB
    from earthkit.regrid.utils.caching import CACHE, SETTINGS

    SYS_DB._clear_index()

    if "tmp_cache" in marks_in_items:
        # ensure these tests use a temporary cache
        SETTINGS["cache-policy"] = "temporary"
        CACHE._settings_changed()
    else:
        SETTINGS["cache-policy"] = "user"
        CACHE._settings_changed()



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import os

import numpy as np
import pytest

from earthkit.regrid import interpolate

PATH = os.path.dirname(__file__)

URL_ROOT = "https://get.ecmwf.int/repository/test-data/earthkit-regrid/test-data"


def simple_download(url, target):
    import requests

    r = requests.get(url, allow_redirects=True)
    r.raise_for_status()
    open(target, "wb").write(r.content)


def get_test_data(filename, subfolder="global_0_360"):
    if not isinstance(filename, list):
        filename = [filename]

    res = []
    for fn in filename:
        d_path = os.path.join(PATH, "data", subfolder)
        os.makedirs(d_path, exist_ok=True)
        f_path = os.path.join(d_path, fn)
        if not os.path.exists(f_path):
            simple_download(url=f"{URL_ROOT}/{subfolder}/{fn}", target=f_path)
        res.append(f_path)

    if len(res) == 1:
        return res[0]
    else:
        return res


@pytest.mark.download
@pytest.mark.tmp_cache
@pytest.mark.parametrize(
    "_kwarg,method",
    [
        ({}, "linear"),
        ({"method": "linear"}, "linear"),
        ({"method": "nearest-neighbour"}, "nearest-neighbour"),
        ({"method": "nn"}, "nearest-neighbour"),
        ({"method": "nearest-neighbor"}, "nearest-neighbour"),
    ],
)
def test_method_kwarg(_kwarg, method):
    f_in, f_out = get_test_data(["in_O32.npz", f"out_O32_10x10_{method}.npz"])

    v_in = np.load(f_in)["arr_0"]
    v_ref = np.load(f_out)["arr_0"]
    v_res = interpolate(v_in, {"grid": "O32"}, {"grid": [10, 10]}, **_kwarg)

    assert v_res.shape == (19, 36), 1
    assert np.allclose(v_res.flatten(), v_ref), 1


def _ll_to_ll(method):
    f_in, f_out = get_test_data(["in_5x5.npz", f"out_5x5_10x10_{method}.npz"])

    v_in = np.load(f_in)["arr_0"]
    v_ref = np.load(f_out)["arr_0"]
    v_res = interpolate(v_in, {"grid": [5, 5]}, {"grid": [10, 10]}, method=method)

    assert v_res.shape == (19, 36), 1
    assert np.allclose(v_res.flatten(), v_ref), 1

    # repeated use
    v_res = interpolate(v_in, {"grid": [5, 5]}, {"grid": [10, 10]}, method=method)

    assert v_res.shape == (19, 36), 1
    assert np.allclose(v_res.flatten(), v_ref), 1


@pytest.mark.download
@pytest.mark.tmp_cache
@pytest.mark.parametrize("method", ["linear", "nearest-neighbour"])
def test_ll_to_ll(method):
    _ll_to_ll(method)

    # f_in, f_out = get_test_data(["in_5x5.npz", f"out_5x5_10x10_{method}.npz"])

    # v_in = np.load(f_in)["arr_0"]
    # v_ref = np.load(f_out)["arr_0"]
    # v_res = interpolate(v_in, {"grid": [5, 5]}, {"grid": [10, 10]}, method=method)

    # assert v_res.shape == (19, 36), 1
    # assert np.allclose(v_res.flatten(), v_ref), 1

    # # repeated use
    # v_res = interpolate(v_in, {"grid": [5, 5]}, {"grid": [10, 10]}, method=method)

    # assert v_res.shape == (19, 36), 1
    # assert np.allclose(v_res.flatten(), v_ref), 1


@pytest.mark.download
@pytest.mark.parametrize("method", ["linear", "nearest-neighbour"])
def test_ll_to_ll_user_cache(method):
    _ll_to_ll(method)

    # f_in, f_out = get_test_data(["in_5x5.npz", f"out_5x5_10x10_{MATRIX_VERSION}.npz"])

    # v_in = np.load(f_in)["arr_0"]
    # v_ref = np.load(f_out)["arr_0"]
    # v_res = interpolate(v_in, {"grid": [5, 5]}, {"grid": [10, 10]})

    # assert v_res.shape == (19, 36), 1
    # assert np.allclose(v_res.flatten(), v_ref), 1

    # # repeated use
    # v_res = interpolate(v_in, {"grid": [5, 5]}, {"grid": [10, 10]})

    # assert v_res.shape == (19, 36), 1
    # assert np.allclose(v_res.flatten(), v_ref), 1


@pytest.mark.download
@pytest.mark.tmp_cache
@pytest.mark.parametrize("method", ["linear", "nearest-neighbour"])
def test_ogg_to_ll(method):
    f_in, f_out = get_test_data(["in_O32.npz", f"out_O32_10x10_{method}.npz"])

    v_in = np.load(f_in)["arr_0"]
    v_ref = np.load(f_out)["arr_0"]
    v_res = interpolate(v_in, {"grid": "O32"}, {"grid": [10, 10]}, method=method)

    assert v_res.shape == (19, 36)
    assert np.allclose(v_res.flatten(), v_ref)


@pytest.mark.download
@pytest.mark.tmp_cache
@pytest.mark.parametrize("method", ["linear", "nearest-neighbour"])
def test_ngg_to_ll(method):
    f_in, f_out = get_test_data(["in_N32.npz", f"out_N32_10x10_{method}.npz"])

    v_in = np.load(f_in)["arr_0"]
    v_ref = np.load(f_out)["arr_0"]
    v_res = interpolate(v_in, {"grid": "N32"}, {"grid": [10, 10]}, method=method)

    assert v_res.shape == (19, 36)
    assert np.allclose(v_res.flatten(), v_ref)


@pytest.mark.download
@pytest.mark.tmp_cache
@pytest.mark.parametrize("method", ["linear", "nearest-neighbour"])
def test_healpix_ring_to_ll(method):
    f_in, f_out = get_test_data(["in_H4_ring.npz", f"out_H4_ring_10x10_{method}.npz"])

    v_in = np.load(f_in)["arr_0"]
    v_ref = np.load(f_out)["arr_0"]
    v_res = interpolate(
        v_in, {"grid": "H4", "ordering": "ring"}, {"grid": [10, 10]}, method=method
    )

    assert v_res.shape == (19, 36)
    assert np.allclose(v_res.flatten(), v_ref)


@pytest.mark.download
@pytest.mark.tmp_cache
@pytest.mark.parametrize("method", ["linear", "nearest-neighbour"])
def test_healpix_nested_to_ll(method):
    f_in, f_out = get_test_data(
        ["in_H4_nested.npz", f"out_H4_nested_10x10_{method}.npz"]
    )

    v_in = np.load(f_in)["arr_0"]
    v_ref = np.load(f_out)["arr_0"]
    v_res = interpolate(
        v_in, {"grid": "H4", "ordering": "nested"}, {"grid": [10, 10]}, method=method
    )

    assert v_res.shape == (19, 36)
    assert np.allclose(v_res.flatten(), v_ref)


@pytest.mark.tmp_cache
def test_unsupported_input_grid() -> None:
    a = np.ones(91 * 180)
    with pytest.raises(ValueError):
        _ = interpolate(a, {"grid": [2.2333424, 2]}, {"grid": [1, 1]}, method="linear")


@pytest.mark.tmp_cache
def test_unsupported_output_grid() -> None:
    a = np.ones(181 * 360)
    with pytest.raises(ValueError):
        _ = interpolate(a, {"grid": [1.11323424, 1]}, {"grid": [5, 5]}, method="linear")



# (C) Copyright 2022 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

# Taken from: https://github.com/michaeljones/sphinx-xref

from docutils import nodes
from sphinx.util import caption_ref_re


def xref(typ, rawtext, text, lineno, inliner, options={}, content=[]):
    title = target = text

    # look if explicit title and target are given with `foo <bar>` syntax
    brace = text.find("<")
    if brace != -1:
        m = caption_ref_re.match(text)
        if m:
            target = m.group(2)
            title = m.group(1)
        else:
            # fallback: everything after '<' is the target
            target = text[brace + 1 :]
            title = text[:brace]

    link = xref.links[target]

    if brace != -1:
        pnode = nodes.reference(target, title, refuri=link[1])
    else:
        pnode = nodes.reference(target, link[0], refuri=link[1])

    return [pnode], []


def get_refs(app):
    xref.links = app.config.xref_links


def setup(app):
    app.add_config_value("xref_links", {}, True)
    app.add_role("xref", xref)
    app.connect("builder-inited", get_refs)



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


def bytes(n):
    u = ["", " KiB", " MiB", " GiB", " TiB", " PiB", "EiB", "ZiB", "YiB"]
    i = 0
    while n >= 1024:
        n /= 1024.0
        i += 1
    return "%g%s" % (int(n * 10 + 0.5) / 10.0, u[i])



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import tempfile


class TmpFile:
    """The TmpFile objects are designed to be used for temporary files.
    It ensures that the file is unlinked when the object is
    out-of-scope (with __del__).

    Parameters
    ----------
    path : str
        Actual path of the file.
    """

    def __init__(self, path: str):
        self.path = path

    def __del__(self):
        self.cleanup()

    def __enter__(self):
        return self.path

    def __exit__(self, *args, **kwargs):
        self.cleanup()

    def cleanup(self):
        if self.path is not None:
            os.unlink(self.path)
        self.path = None


def temp_file(extension=".tmp") -> TmpFile:
    """Create a temporary file with the given extension .

    Parameters
    ----------
    extension : str, optional
        By default ".tmp"

    Returns
    -------
    TmpFile
    """
    fd, path = tempfile.mkstemp(suffix=extension)
    os.close(fd)
    return TmpFile(path)


class TmpDirectory(tempfile.TemporaryDirectory):
    def __init__(self, *args, **kwargs):
        d = kwargs.get("dir", None)
        if d == "":
            raise ValueError("TmpDirectory: dir option cannot be an empty str!")
        if d is not None and not os.path.exists(d):
            os.makedirs(d)
        super().__init__(*args, **kwargs)

    @property
    def path(self):
        return self.name


def temp_directory(dir=None):
    return TmpDirectory(dir=dir)



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import re

LOG = logging.getLogger(__name__)

FULL_GLOBE = 360.0
FULL_GLOBE_EPS = 1e-8
DEGREE_EPS = 1e-8

HEALPIX_PATTERN = re.compile(r"[Hh]\d+")
RGG_PATTERN = re.compile(r"[OoNn]\d+")


# NOTE: this is a temporary code until the full gridspec
# implementation is available via earthkit-geo.


def same_coord(x, y, eps=DEGREE_EPS):
    return abs(x - y) < eps


class GridSpec(dict):
    DEFAULTS = {
        "i_scans_negatively": 0,
        "j_points_consecutive": 0,
        "j_scans_positively": 0,
    }
    DEFAULT_AREA = [90, 0, -90, 360]

    COMPARE_KEYS = {
        "type",
        "grid",
        "i_scans_negatively",
        "j_points_consecutive",
        "j_scans_positively",
    }

    def __init__(self, gs):
        self._global_ew = None
        self._global_ns = None
        super().__init__(gs)

    @staticmethod
    def from_dict(d):
        gs = dict(GridSpec.DEFAULTS)
        gs.update(d)
        t_name, t = GridSpec._infer_spec_type(gs)
        if t is None:
            raise ValueError(f"Unsupported gridspec={d}")

        gs["type"] = t_name
        return t(gs)

    @staticmethod
    def _infer_spec_type(spec):
        spec_type = spec.get("type", None)
        if spec_type is None:
            grid = spec["grid"]
            for k, gs in GRIDSPEC_TYPES.items():
                if gs.type_match(grid):
                    return k, gs

        return spec_type, GRIDSPEC_TYPES.get(spec_type, None)

    def __eq__(self, o):
        # print(f"__eq__ self={self} o={o}")
        if self.get("type", "") != o.get("type", ""):
            return False

        for k in self.COMPARE_KEYS:
            if str(self[k]) != str(o[k]):
                return False

        if "shape" in self and "shape" in o:
            if self["shape"] != o["shape"]:
                return False
        return True

    @staticmethod
    def same_area(area1, area2, eps=DEGREE_EPS):
        if len(area1) == len(area2):
            return all(same_coord(v1, v2, eps=eps) for v1, v2 in zip(area1, area2))
        return False

    def has_default_area(self):
        return self.same_area(self["area"], self.DEFAULT_AREA)

    @staticmethod
    def normalise_lon(lon, minimum):
        while lon < minimum:
            lon += FULL_GLOBE
        while lon >= minimum + FULL_GLOBE:
            lon -= FULL_GLOBE
        return lon

    def is_global(self):
        return self.is_global_ew() and self.is_global_ns()

    def is_global_ew(self):
        if self._global_ew is not None:
            return self._global_ew
        else:
            raise NotADirectoryError

    def is_global_ns(self):
        if self._global_ns is not None:
            return self._global_ns
        else:
            raise NotADirectoryError


class LLGridSpec(GridSpec):
    GLOBAL_AREAS = {
        (0.1, 0.1): [90, 0, -90, 359.9],
        (0.125, 0.125): [90, 0, -90, 359.875],
        (0.15, 0.15): [90, 0, -90, 359.85],
        (0.2, 0.2): [90, 0, -90, 359.8],
        (0.25, 0.25): [90, 0, -90, 359.75],
        (0.3, 0.3): [90, 0, -90, 359.7],
        (0.4, 0.4): [90, 0, -90, 359.6],
        (0.5, 0.5): [90, 0, -90, 359.5],
        (0.6, 0.6): [90, 0, -90, 359.4],
        (0.7, 0.7): [89.6, 0, -89.6, 359.8],
        (0.75, 0.75): [90, 0, -90, 359.25],
        (0.8, 0.8): [89.6, 0, -89.6, 359.2],
        (0.9, 0.9): [90, 0, -90, 359.1],
        (1, 1): [90, 0, -90, 359],
        (1.2, 1.2): [90, 0, -90, 358.8],
        (1.25, 1.25): [90, 0, -90, 358.75],
        (1.4, 1.4): [89.6, 0, -89.6, 359.8],
        (1.5, 1.5): [90, 0, -90, 358.5],
        (1.6, 1.6): [89.6, 0, -89.6, 358.4],
        (1.8, 1.8): [90, 0, -90, 358.2],
        (2, 2): [90, 0, -90, 358],
        (2.5, 2.5): [90, 0, -90, 357.5],
        (5, 5): [90, 0, -90, 355],
        (10, 10): [90, 0, -90, 350],
    }

    def __init__(self, gs):
        super().__init__(gs)

        if "global" not in gs and "area" not in gs:
            self["global"] = 1
            self["area"] = LLGridSpec.GLOBAL_AREAS.get(
                (self["grid"][0], self["grid"][1]), self.DEFAULT_AREA
            )

        self.setdefault("area", self.DEFAULT_AREA)

        if self.get("global", 0) or self.has_default_area():
            self._global_ew = True
            self._global_ns = True

    def __eq__(self, o):
        if not super().__eq__(o):
            return False

        if self.same_area(self["area"], o["area"]):
            return True

        # sanity check: north and south must be the same
        if not same_coord(self.north, o.north) or not same_coord(self.south, o.south):
            return False

        # west-east
        if self.is_global_ew() and o.is_global_ew():
            west = self.normalise_lon(self.west, 0)
            west_o = self.normalise_lon(o.west, 0)
            if same_coord(west, west_o):
                return True

        # TODO: add code for non global grids
        return False

    @property
    def west(self):
        return self["area"][1]

    @property
    def east(self):
        return self["area"][3]

    @property
    def north(self):
        return self["area"][0]

    @property
    def south(self):
        return self["area"][2]

    @property
    def dx(self):
        return abs(self["grid"][0])

    def is_global_ew(self):
        if self._global_ew is None:
            west = self.normalise_lon(self.west, 0)
            east = self.normalise_lon(self.east, 0)
            self._global_ew = False
            if east < west:
                east += FULL_GLOBE
            if abs(east - west) < FULL_GLOBE_EPS:
                self._global_ew = True
            elif abs(east - west - FULL_GLOBE) < FULL_GLOBE_EPS:
                self._global_ew = True
            elif abs(FULL_GLOBE - (east - west) - self.dx) < FULL_GLOBE_EPS:
                self._global_ew = True
        return self._global_ew

    @staticmethod
    def type_match(grid):
        if isinstance(grid, list) and len(grid) == 2:
            return True
        return False


class ReducedGGGridSpec(GridSpec):
    def __init__(self, gs):
        super().__init__(gs)

        self.setdefault("area", self.DEFAULT_AREA)
        if self.get("global", 0) or self.has_default_area():
            self._global_ew = True
            self._global_ns = True

        self._octahedral = None
        self._N = None
        self._eps = 0.12

    def __eq__(self, o):
        if not super().__eq__(o):
            return False

        if self.same_area(self["area"], o["area"], eps=self._eps):
            return True

        # check if west the same for global grids
        if self.is_global() and o.is_global():
            west = self.normalise_lon(self.west, 0)
            west_o = self.normalise_lon(o.west, 0)
            if same_coord(west, west_o, eps=self._eps):
                return True

        # TODO: add code for non global grids
        return False

    @property
    def west(self):
        return self["area"][1]

    @property
    def east(self):
        return self["area"][3]

    @property
    def dx(self):
        if self.octahedral:
            return FULL_GLOBE / (4 * self.N + 16)
        else:
            return FULL_GLOBE / (4 * self.N)

    def is_global_ew(self):
        if self._global_ew is None:
            west = self.normalise_lon(self.west, 0)
            east = self.normalise_lon(self.east, 0)
            self._global_ew = False
            if east < west:
                east += FULL_GLOBE
            if abs(east - west) < FULL_GLOBE_EPS:
                self._global_ew = True
            elif abs(east - west - FULL_GLOBE) < FULL_GLOBE_EPS:
                self._global_ew = True
            elif abs(FULL_GLOBE - (east - west) - self.dx) < FULL_GLOBE_EPS:
                self._global_ew = True
        return self._global_ew

    def is_global_ns(self):
        if self._global_ns is None:
            self._global_ns = True
        return self._global_ns

    def _inspect_grid(self):
        if self._N is None or self._octahedral is None:
            grid = self["grid"]
            octahedral = self.get("octahedral", 0)
            if not isinstance(grid, str) or not RGG_PATTERN.match(grid):
                raise ValueError(f"Invalid {grid=}")
            try:
                if grid[0] == "N":
                    N = int(grid[1:])
                    octahedral = 0
                elif grid[0] == "O":
                    N = int(grid[1:])
                    octahedral = 1
            except Exception as e:
                raise ValueError(f"Invalid {grid=} {e}")
            if N < 1 or N > 1000000:
                raise ValueError(f"Invalid {N=}")
            self._N = N
            self._octahedral = octahedral

    @property
    def N(self):
        if self._N is None:
            self._inspect_grid()
        return self._N

    @property
    def octahedral(self):
        if self._octahedral is None:
            self._inspect_grid()
        return self._octahedral

    @staticmethod
    def type_match(grid):
        if isinstance(grid, str):
            if RGG_PATTERN.match(grid):
                return True
        return False


class HealpixGridSpec(GridSpec):
    def __init__(self, gs):
        super().__init__(gs)

        self._global_ew = True
        self._global_ns = True
        self.setdefault("area", self.DEFAULT_AREA)
        self._ordering = None
        self._N = None

    @property
    def west(self):
        return self["area"][1]

    @property
    def east(self):
        return self["area"][3]

    @property
    def dx(self):
        if self.octahedral:
            return FULL_GLOBE / (4 * self.N + 16)
        else:
            return FULL_GLOBE / (4 * self.N)

    def __eq__(self, o):
        if not super().__eq__(o):
            return False

        return self.ordering == o.ordering

    def _inspect_grid(self):
        if self._N is None or self._ordering is None:
            grid = self["grid"]
            if not self.type_match(grid):
                raise ValueError(f"Invalid healpix {grid=}")

            try:
                N = int(grid[1:])
            except Exception:
                raise ValueError(f"Invalid healpix number in {grid=}")

            ordering = self.get("ordering", "ring")
            if ordering not in ["ring", "nested"]:
                raise ValueError(f"Invalid {ordering=}, must be 'ring' or 'nested'")

            if N < 1 or N > 1000000:
                raise ValueError(f"Invalid healpix number in {grid=}")
            self._N = N
            self._ordering = ordering

    @property
    def N(self):
        if self._N is None:
            self._inspect_grid()
        return self._N

    @property
    def ordering(self):
        if self._ordering is None:
            self._inspect_grid()
        return self._ordering

    @staticmethod
    def type_match(grid):
        if isinstance(grid, str):
            return HEALPIX_PATTERN.match(grid)
        return False


GRIDSPEC_TYPES = {
    "regular_ll": LLGridSpec,
    "reduced_gg": ReducedGGGridSpec,
    "healpix": HealpixGridSpec,
}



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import json
import logging
import os
from abc import ABCMeta, abstractmethod
from functools import lru_cache

from scipy.sparse import load_npz

from earthkit.regrid.gridspec import GridSpec
from earthkit.regrid.utils import no_progress_bar
from earthkit.regrid.utils.download import download_and_cache

LOG = logging.getLogger(__name__)

VERSION = 1

_SYSTEM_URL = "https://get.ecmwf.int/repository/earthkit/regrid/db/1/"
_INDEX_FILENAME = "index.json"
_INDEX_SHA_FILENAME = "index.json.sha256"
_INDEX_GZ_FILENAME = "index.json.gz"


_METHOD_ALIAS = {"nearest-neighbour": ("nn", "nearest-neighbor")}


def make_sha(data):
    import hashlib

    m = hashlib.sha256()
    if isinstance(data, str):
        m.update(data.encode("utf-8"))
    else:
        m.update(json.dumps(data, sort_keys=True).encode("utf-8"))
    return m.hexdigest()


class MatrixAccessor(metaclass=ABCMeta):
    @abstractmethod
    def path(self):
        pass

    @abstractmethod
    def is_local(self):
        pass

    @abstractmethod
    def index_path(self):
        pass

    @abstractmethod
    def matrix_path(self, name):
        pass


class UrlAccessor(MatrixAccessor):
    def __init__(self, url):
        self._url = url

    def path(self):
        return self._url

    def is_local(self):
        False

    @lru_cache
    def index_path(self):
        from earthkit.regrid.utils.caching import cache_file

        url = os.path.join(self._url, _INDEX_FILENAME)

        def _force(args, path, owner_data):
            """Decide if the index file should be downloaded and cached again."""
            LOG.info(f"UrlAccessor.index_path: cached entry {owner_data=}")
            if owner_data is None:
                return True
            local_sha = owner_data.get("sha256", None)
            LOG.info(f"UrlAccessor.index_path: local checksum={local_sha}")
            if local_sha is None:
                return True

            remote_sha = self._remote_sha()
            if local_sha != remote_sha:
                LOG.info(
                    (
                        f"UrlAccessor.index_path: remote checksum={remote_sha} differs from "
                        "local checksum. Downloading new index file"
                    )
                )
                return True
            else:
                LOG.info(
                    (
                        f"UrlAccessor.index_path: remote checksum={remote_sha} is the same as "
                        "local checksum. Use cached index file."
                    )
                )
                return False

        def _create(target, args):
            """Download and cache gzipped index file, uncompress it and generate
            checksum from contents
            """
            path_gz = self._gzip_file()
            LOG.info(f"UrlAccessor.index_path: uncompress gzipped index file={path_gz}")
            import gzip

            with gzip.open(path_gz, "rb") as f:
                data = f.read()
                with open(target, "wb") as f_out:
                    f_out.write(data)

            with open(target, "r") as f:
                data = f.read()
                sha = make_sha(data)

            # the returned data will be stored in the cache in the entry's owner_data
            LOG.info(f"UrlAccessor.index_path: index file checksum={sha}")
            return {"sha256": sha}

        path = cache_file(
            "regrid",
            _create,
            (url,),
            force=_force,
            extension=".cache",
        )

        LOG.info(f"UrlAccessor.index_path: will use index file={path}")
        return path

    def _remote_sha(self):
        try:
            url = os.path.join(self._url, _INDEX_SHA_FILENAME)
            path = download_and_cache(
                url,
                owner="url",
                verify=True,
                force=True,
                chunk_size=1024 * 1024,
                http_headers=None,
                update_if_out_of_date=True,
                progress_bar=no_progress_bar,
                maximum_retries=0,
                retry_after=10,
            )
        except Exception:
            LOG.error(f"Could not download index checksum file={url}")
            raise

        with open(path, "r") as f:
            sha = f.read().strip()
        return sha

    def _gzip_file(self):
        try:
            url = os.path.join(self._url, _INDEX_GZ_FILENAME)
            path = download_and_cache(
                url,
                owner="url",
                verify=True,
                force=True,
                chunk_size=1024 * 1024,
                http_headers=None,
                update_if_out_of_date=True,
                progress_bar=no_progress_bar,
                maximum_retries=0,
                retry_after=10,
            )
        except Exception:
            LOG.error(f"Could not download index file={url}")
            raise

        return path

    def matrix_path(self, name):
        try:
            url = os.path.join(self._url, name)
            path = download_and_cache(
                url,
                owner="url",
                verify=True,
                force=None,
                chunk_size=1024 * 1024,
                http_headers=None,
                update_if_out_of_date=False,
                maximum_retries=5,
                retry_after=10,
            )
        except Exception:
            LOG.error(f"Could not download matrix file={url}")
            raise

        return path


class LocalAccessor(MatrixAccessor):
    def __init__(self, path):
        self._path = path

    def path(self):
        return self._path

    def is_local(self):
        True

    def index_path(self):
        return os.path.join(self._path, _INDEX_FILENAME)

    def matrix_path(self, name):
        return os.path.join(self._path, name)


class SystemAccessor(UrlAccessor):
    def __init__(self):
        super().__init__(_SYSTEM_URL)


class MatrixIndex(dict):
    def load(self, path):
        with open(path, "r") as f:
            index = json.load(f)
            version = index.get("version", None)
            if version != VERSION:
                raise ValueError(
                    f"Invalid index file version: expected {VERSION}, got {version}"
                )

            for name, entry in index["matrix"].items():
                # it is possible that the inventory is already updated with new
                # a gridspecs type, but a given earthkit-regrid version is not
                # yet supporting it. In this case loading the index should not crash.
                try:
                    in_gs = GridSpec.from_dict(entry["input"])
                    out_gs = GridSpec.from_dict(entry["output"])
                    raw = entry
                    entry = dict(**entry)
                    entry["input"] = in_gs
                    entry["output"] = out_gs
                    entry["_name"] = name
                    entry["_raw"] = raw
                    # print(f"{in_gs=}")
                    # print(f"{out_gs=}")
                    self[name] = entry
                except Exception:
                    pass

    @staticmethod
    def make_interpolation_uid(item):
        inter = item["interpolation"]
        method = inter["method"]
        if set(inter.keys()) == {"method", "engine", "version"}:
            uid = method
        else:
            uid = make_sha(inter)
        return uid

    @staticmethod
    def matrix_dir_name(item):
        inter = item["interpolation"]
        engine = inter["engine"]
        version = inter["version"]
        uid = inter.get("_uid", inter["method"])
        return f"{engine}_{version}_{uid}"

    @staticmethod
    def matrix_path(item):
        return os.path.join(MatrixIndex.matrix_dir_name(item), item["_name"] + ".npz")

    def find(self, gridspec_in, gridspec_out, method):
        gridspec_in = GridSpec.from_dict(gridspec_in)
        gridspec_out = GridSpec.from_dict(gridspec_out)

        # print(f"{gridspec_in=} {gridspec_out=} {method=}")
        if gridspec_in is None or gridspec_out is None:
            return None

        for _, entry in self.items():
            if self.match(entry, gridspec_in, gridspec_out, method):
                return entry
        return None

    @staticmethod
    def match(item, gs_in, gs_out, method):
        if (
            item["interpolation"]["method"] == method
            and item["input"] == gs_in
            and item["output"] == gs_out
        ):
            return True
        return False

    @staticmethod
    def matrix_filename(item):
        return item["_name"] + ".npz"

    def subset(self, filters, fail_on_missing=True, raw=False):
        res = MatrixIndex()

        for i, item in enumerate(filters):
            gs_in = item["input"]
            gs_out = item["output"]
            method = item.get("method", "linear")
            LOG.info(f"ITEM[{i}]: {gs_in=} {gs_out=} {method=}")
            entry = self.find(gs_in, gs_out, method)
            if entry is not None:
                LOG.info("  found DB entry:" + entry["_name"])
                if raw:
                    res[entry["_name"]] = entry["_raw"]
                else:
                    res[entry["_name"]] = entry
            else:
                if fail_on_missing:
                    raise ValueError("No DB entry found!")
                else:
                    LOG.warning("  no DB entry found!")
        return res

    def to_raw(self):
        res = dict(version=VERSION, matrix={})
        for _, entry in self.items():
            res["matrix"][entry["_name"]] = entry["_raw"]
        return res


class MatrixDb:
    def __init__(self, accessor):
        self._index = None
        self._accessor = accessor

    @property
    def index(self):
        if self._index is None:
            self._load_index()
        return self._index

    def _load_index(self):
        self._index = MatrixIndex()
        path = self._accessor.index_path()
        self._index.load(path)

    def _clear_index(self):
        """For testing only"""
        self._index = None

    def _method_alias(self, method):
        for k, v in _METHOD_ALIAS.items():
            if method in v:
                return k
        return method

    def find(
        self,
        gridspec_in,
        gridspec_out,
        method,
        **kwargs,
    ):
        entry = self.find_entry(gridspec_in, gridspec_out, method)

        if entry is not None:
            z = self.load_matrix(entry)
            return z, entry["output"]["shape"]
        return None, None

    def find_entry(self, gridspec_in, gridspec_out, method):
        method = self._method_alias(method)
        return self.index.find(gridspec_in, gridspec_out, method)

    def load_matrix(self, entry):
        path = self._matrix_fs_path(entry)
        z = load_npz(path)
        return z

    def _matrix_index_filename(self, entry):
        return self.index.matrix_filename(entry)

    def _matrix_index_path(self, entry):
        return self.index.matrix_path(entry)

    def _matrix_fs_path(self, entry):
        return self._accessor.matrix_path(self._matrix_index_path(entry))

    def subset_index(self, filters, **kwargs):
        return self.index.subset(filters, **kwargs)

    def copy_matrix_file(self, entry, out_dir, exist_ok=False, dry_run=False):
        import shutil

        matrix_index_path = self._matrix_index_path(entry)
        src_file = self._matrix_fs_path(entry)
        target_file = os.path.join(out_dir, matrix_index_path)

        if not exist_ok and os.path.exists(target_file):
            if not dry_run:
                raise FileExistsError(f"target file already exists! {target_file}")
            else:
                LOG.warn("target file already exists! {target_file}")

        target_dir = os.path.dirname(target_file)

        if not dry_run:
            os.makedirs(target_dir, exist_ok=True)
            shutil.copyfile(src_file, target_file)

        return target_file

    def index_file_path(self):
        return self._accessor.index_path()

    def matrix_source(self):
        return self._accessor.path()

    @staticmethod
    def from_path(path):
        return MatrixDb(LocalAccessor(path))

    def __len__(self):
        return len(self.index)


SYS_DB = MatrixDb(SystemAccessor())
DB_LIST = [SYS_DB]


def add_matrix_source(path):
    global DB_LIST
    for item in DB_LIST[1:]:
        if item.matrix_source() == path:
            return item
    db = MatrixDb.from_path(path)
    DB_LIST.append(db)
    return db


def find(*args, matrix_source=None, **kwargs):
    if matrix_source is None:
        return SYS_DB.find(*args, **kwargs)
    else:
        db = add_matrix_source(matrix_source)
        return db.find(*args, **kwargs)



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os
import re
from functools import cache

LOG = logging.getLogger(__name__)

GRIB_DIR = "grib"
MIR_PATH = os.path.expanduser("~/build/mir/release/bin/mir")
MIR_INTERPOLATE_OPTION = "interpolation"


def download_index_file(target):
    import requests

    from earthkit.regrid.db import _INDEX_GZ_FILENAME, _SYSTEM_URL

    path = os.path.join(_SYSTEM_URL, _INDEX_GZ_FILENAME)
    requests.get(path)
    r = requests.get(path, allow_redirects=True)
    target_gz = target + ".gz"
    open(target_gz, "wb").write(r.content)
    import gzip

    with gzip.open(target_gz, "rb") as f:
        data = f.read()
        with open(target, "wb") as f_out:
            f_out.write(data)


def adjust_grid_name(name):
    d = {"n": "N", "o": "O", "f": "F", "/": "x"}
    for k, v in d.items():
        name = name.replace(k, v)
    return name


@cache
def get_mir_version():
    import subprocess

    res = subprocess.run(
        [MIR_PATH, "--version", "x", "x"],
        stdout=subprocess.PIPE,
    )
    for line in res.stdout.decode("utf-8").split("\n"):
        if line.lower().startswith("mir"):
            return line.split(" ")[1]


def get_grib_file(grid, grib_file):
    if not os.path.exists(grib_file):
        from earthkit.data import from_source

        _class = "od"
        expver = "1"
        if grid in ["N2560", "O2560"]:
            _class = "rd"
            expver = "i4ql"

        ds = from_source(
            "mars",
            {
                "levtype": "sfc",
                "param": "2t",
                "grid": grid,
                "class": _class,
                "expver": expver,
                "type": "fc",
                "time": "0",
            },
        )
        ds.save(grib_file)


def create_matrix_files(
    matrix_dir,
    src_grid_label,
    target_grid_label,
    target_grid,
    options,
    grib_file,
    index_file,
    add_to_index=True,
    delete_tmp_json=False,
):
    # generate interpolation matrix
    if options:
        kwargs = " ".join([f"--{k}={v}" for k, v in options.items()])
    else:
        kwargs = ""

    matrix_json = f"{target_grid_label}-{src_grid_label}"
    if MIR_INTERPOLATE_OPTION in options:
        matrix_json += "-" + options[MIR_INTERPOLATE_OPTION]
    matrix_json = os.path.join(matrix_dir, f"{matrix_json}.json")

    if not os.path.exists(matrix_json):
        cmd = (
            f"{MIR_PATH} --grid={target_grid} {kwargs} --dump-weights-info={matrix_json}"
            f" {grib_file} /dev/null"
        )
        LOG.debug(f" {cmd=}")
        os.system(cmd)

    if not os.path.exists(matrix_json):
        print(f"{matrix_json} does not exist!")

    if add_to_index:
        # process matrix and add it to index json file
        from earthkit.regrid.utils.matrix import make_matrix

        make_matrix(
            matrix_json,
            matrix_dir,
            index_file=index_file,
            global_input=True,
            global_output=True,
        )

    if delete_tmp_json:
        os.remove(matrix_json)


def make_matrix(
    src_grid,
    target_grid,
    method,
    matrix_dir,
    index_file=None,
    download_index=False,
    delete_tmp_json=False,
):
    LOG.debug(f"{src_grid=} {target_grid=} {matrix_dir=} {index_file=}")

    options = {}

    src_grid_label = src_grid
    target_grid_label = target_grid

    options[MIR_INTERPOLATE_OPTION] = method

    if not re.match(r"[Hh]\d+_[A-z]+", src_grid):
        #     if src_grid.endswith("_nested"):
        #         options[MIR_INTERPOLATE_OPTION] = "nearest-neighbor"
        # else:
        src_grid_label = adjust_grid_name(src_grid)

    if not re.match(r"[Hh]\d+_[A-z]+", target_grid):
        target_grid_label = adjust_grid_name(target_grid)

    os.makedirs(matrix_dir, exist_ok=True)
    if index_file is None:
        index_file = os.path.join(matrix_dir, "index.json")

    assert os.path.exists(GRIB_DIR)
    assert os.path.exists(matrix_dir)
    assert index_file is not None

    if download_index:
        download_index_file(index_file)
        assert os.path.exists(index_file)

    grib_file = os.path.join(GRIB_DIR, f"{src_grid_label}.grib")
    get_grib_file(src_grid, grib_file)

    # version = get_mir_version()

    create_matrix_files(
        matrix_dir,
        src_grid_label,
        target_grid_label,
        target_grid,
        options,
        grib_file,
        index_file,
        add_to_index=True,
        delete_tmp_json=delete_tmp_json,
    )



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os
import sys

here = os.path.dirname(__file__)
sys.path.insert(0, os.path.dirname(here))

from utils.grid import make_grid_id  # noqa
from utils.matrix import make_matrix  # noqa

LOG = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)

"""
FOR MAINTENANCE ONLY!!!
-----------------------

This script generates a matrix inventory resulting in an index.json file and one
directory per interpolation method with the name of

{engine}_{version}_{method}

where
- engine is the regrid engine,
- version is the regrid cache file version
- method is the interpolation method.

E.g. "mir_16_linear"

The matrices are generated by calling mir on the input grib data stored in
folder "grib" (not part of the git repo). When the data file is not present
an attempt is made to download it from MARS using earthkit-data.

The generated data can be a whole new inventory or just a delta to the existing one.

Uploading all the data to the inventory are done by a separate script.
"""

# shorthand for grid definition:
# ll : 1 - > 1/1
# O_rgg: 32 -> O32
# N_rgg: 32 -> N32
# H_ring: 32 -> H32_ring
# H_nested: 32 -> H32_nested

gg_res = [1280, 1024, 640, 512, 400, 320, 256, 200, 160, 128, 96, 80, 64, 48, 32]

in_grids_ll = []
in_grids_O_gg = [2560] + gg_res
in_grids_N_gg = gg_res

in_grids = {
    "ll": [],
    # "O_rgg": [2560] + gg_res,
    # "N_rgg": gg_res,
    # "H_ring": [1024, 512, 256, 128, 64, 32, 16, 8, 4, 2],
    "H_nested": [1024, 512, 256, 128, 64, 32, 16, 8, 4, 2],
}
out_grids = {
    "ll": [
        0.1,
        0.125,
        0.15,
        0.2,
        0.25,
        0.3,
        0.4,
        0.5,
        0.6,
        0.7,
        0.75,
        0.8,
        0.9,
        1,
        1.2,
        1.25,
        1.4,
        1.5,
        1.6,
        1.8,
        2,
        2.5,
        5,
        10,
    ]
}

# in_grids = {
#     "H_ring": [8],
#     "H_nested": [8],
# }

# out_grids = {
#     "ll": [5],
# }

build_dir = "_build_20240325"

# extra = [["0.25x0.25", "N320"], ["O1280", "N320"], ["5x5", "10x10"]]
extra = []

in_grids = make_grid_id(in_grids)
out_grids = make_grid_id(out_grids)
LOG.debug(f"{in_grids=}")
LOG.debug(f"{out_grids=}")

index_file = os.path.join(build_dir, "index.json")

for method in ["linear", "nn"]:
    matrix_dir = os.path.join(build_dir, f"matrices_{method}")

    for g_in in in_grids:
        for g_out in out_grids:
            if g_in != g_out:
                make_matrix(
                    g_in,
                    g_out,
                    method,
                    matrix_dir,
                    index_file=index_file,
                    download_index=False,
                    delete_tmp_json=False,
                )

    for g_in, g_out in extra:
        make_matrix(
            g_in,
            g_out,
            method,
            matrix_dir,
            index_file=index_file,
            download_index=False,
            delete_tmp_json=False,
        )



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

try:
    # There is a bug in tqdm that expects ipywidgets
    # to be installed if running in a notebook
    import ipywidgets  # noqa F401
    from tqdm.auto import tqdm  # noqa F401
except ImportError:
    from tqdm import tqdm  # noqa F401


def progress_bar(*, total=None, iterable=None, initial=0, desc=None):
    return tqdm(
        iterable=iterable,
        total=total,
        initial=initial,
        unit_scale=True,
        unit_divisor=1024,
        unit="B",
        disable=False,
        leave=False,
        desc=desc,
        # dynamic_ncols=True, # make this the default?
    )


class NoBar:
    def __init__(self, *args, **kwargs):
        pass

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        pass

    def update(self, *args, **kwargs):
        pass

    def close(self, *args, **kwargs):
        pass


def no_progress_bar(total, initial=0, desc=None):
    return NoBar(
        total=total,
        initial=initial,
        unit_scale=True,
        unit_divisor=1024,
        unit="B",
        disable=False,
        leave=False,
        desc=desc,
    )



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


def make_ll(g):
    return f"{g}/{g}"


def make_O_rgg(g):
    return f"O{g}"


def make_N_rgg(g):
    return f"N{g}"


def make_H_ring(g):
    return f"H{g}_ring"


def make_H_nested(g):
    return f"H{g}_nested"


makers = {
    k: globals()[f"make_{k}"] for k in ["ll", "O_rgg", "N_rgg", "H_ring", "H_nested"]
}


def make_grid_id(grids):
    r = []
    for k, v in grids.items():
        m = makers[k]
        for x in v:
            r.append(m(x))
    return r



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import os

import numpy as np
import pytest

from earthkit.regrid import interpolate
from earthkit.regrid.db import add_matrix_source

DB_PATH = os.path.join(os.path.dirname(__file__), "data", "local", "db")
DATA_PATH = os.path.join(os.path.dirname(__file__), "data", "local")


def file_in_testdir(filename):
    return os.path.join(DATA_PATH, filename)


def test_local_index():
    DB = add_matrix_source(DB_PATH)
    # we have an extra unsupported entry in the index file. We have
    # to be sure the DB is loaded correctly bypassing the unsupported
    # entry.
    import json

    method = "linear"

    index_path = DB.index_file_path()
    with open(index_path, "r") as f:
        d = json.load(f)
        assert len(d["matrix"]) == 11

    assert len(DB) == 10

    # r = DB.find_entry({"grid": [5, 5]}, {"grid": [10, 10]}, method)
    # assert r

    r = DB.find_entry({"grid": "O32"}, {"grid": [10, 10]}, method)
    assert r

    r = DB.find_entry({"grid": "N32"}, {"grid": [10, 10]}, method)
    assert r

    r = DB.find_entry({"grid": "O64"}, {"grid": [10, 10]}, method)
    assert r is None


@pytest.mark.parametrize("method", ["linear", "nearest-neighbour"])
def test_local_ll_to_ll(method):
    v_in = np.load(file_in_testdir("in_5x5.npz"))["arr_0"]
    v_ref = np.load(file_in_testdir(f"out_5x5_10x10_{method}.npz"))["arr_0"]
    v_res = interpolate(
        v_in, {"grid": [5, 5]}, {"grid": [10, 10]}, matrix_source=DB_PATH, method=method
    )

    assert v_res.shape == (19, 36)
    assert np.allclose(v_res.flatten(), v_ref)


@pytest.mark.parametrize("method", ["linear", "nearest-neighbour"])
def test_local_ogg_to_ll(method):
    v_in = np.load(file_in_testdir("in_O32.npz"))["arr_0"]
    v_ref = np.load(file_in_testdir(f"out_O32_10x10_{method}.npz"))["arr_0"]
    v_res = interpolate(
        v_in, {"grid": "O32"}, {"grid": [10, 10]}, matrix_source=DB_PATH, method=method
    )

    assert v_res.shape == (19, 36)
    assert np.allclose(v_res.flatten(), v_ref)


@pytest.mark.parametrize("method", ["linear", "nearest-neighbour"])
def test_local_ngg_to_ll(method):
    v_in = np.load(file_in_testdir("in_N32.npz"))["arr_0"]
    v_ref = np.load(file_in_testdir(f"out_N32_10x10_{method}.npz"))["arr_0"]
    v_res = interpolate(
        v_in,
        {"grid": "N32"},
        {"grid": [10, 10]},
        matrix_source=DB_PATH,
        method=method,
    )

    assert v_res.shape == (19, 36)
    assert np.allclose(v_res.flatten(), v_ref)


@pytest.mark.parametrize("method", ["linear", "nearest-neighbour"])
def test_local_healpix_ring_to_ll(method):
    v_in = np.load(file_in_testdir("in_H4_ring.npz"))["arr_0"]
    v_ref = np.load(file_in_testdir(f"out_H4_ring_10x10_{method}.npz"))["arr_0"]
    v_res = interpolate(
        v_in,
        {"grid": "H4", "ordering": "ring"},
        {"grid": [10, 10]},
        matrix_source=DB_PATH,
        method=method,
    )

    assert v_res.shape == (19, 36)
    assert np.allclose(v_res.flatten(), v_ref)


@pytest.mark.parametrize("method", ["linear", "nearest-neighbour"])
def test_local_healpix_nested_to_ll(method):
    v_in = np.load(file_in_testdir("in_H4_nested.npz"))["arr_0"]
    v_ref = np.load(file_in_testdir(f"out_H4_nested_10x10_{method}.npz"))["arr_0"]
    v_res = interpolate(
        v_in,
        {"grid": "H4", "ordering": "nested"},
        {"grid": [10, 10]},
        matrix_source=DB_PATH,
        method=method,
    )

    assert v_res.shape == (19, 36)
    assert np.allclose(v_res.flatten(), v_ref)


@pytest.mark.parametrize(
    "gs_in, gs_out",
    [
        ({"grid": [5, 5]}, {"grid": [10, 10]}),
        ({"grid": [5, 5], "area": [90, 0, -90, 360]}, {"grid": [10, 10]}),
        ({"grid": [5, 5], "area": [90, 0, -90, 355]}, {"grid": [10, 10]}),
        ({"grid": [5, 5]}, {"grid": [10, 10], "area": [90, 0, -90, 360]}),
        ({"grid": [5, 5]}, {"grid": [10, 10], "area": [90, 0, -90, 350]}),
        (
            {"grid": [5, 5], "area": [90, 0, -90, 360]},
            {"grid": [10, 10], "area": [90, 0, -90, 360]},
        ),
        ({"grid": [5, 5], "shape": [37, 72]}, {"grid": [10, 10], "shape": [19, 36]}),
        ({"grid": [5, 5], "shape": [37, 72]}, {"grid": [10, 10]}),
        ({"grid": [5, 5]}, {"grid": [10, 10], "shape": [19, 36]}),
        ({"grid": "O32"}, {"grid": [10, 10]}),
        ({"grid": "O32", "shape": [5248]}, {"grid": [10, 10]}),
        ({"grid": "O32", "area": [90, 0, -90, 360]}, {"grid": [10, 10]}),
        ({"grid": "O32", "area": [87.8638, 0, -87.8638, 357.5]}, {"grid": [10, 10]}),
        ({"grid": "O32", "area": [87.8638, 0.01, -87.8638, 357.5]}, {"grid": [10, 10]}),
        ({"grid": "O32", "global": 1}, {"grid": [10, 10]}),
        (
            {"grid": "O32", "global": 1, "area": [87.8638, 0, -87.8638, 357.5]},
            {"grid": [10, 10]},
        ),
        ({"grid": "N32"}, {"grid": [10, 10]}),
        ({"grid": "N32", "shape": [6114]}, {"grid": [10, 10]}),
        ({"grid": "N32", "area": [90, 0, -90, 360]}, {"grid": [10, 10]}),
        ({"grid": "N32", "area": [87.8638, 0, -87.8638, 357.188]}, {"grid": [10, 10]}),
        ({"grid": "N32", "area": [87.8638, 0, -87.8638, 357.189]}, {"grid": [10, 10]}),
        (
            {"grid": "N32", "area": [87.8638, 0.01, -87.8638, 357.188]},
            {"grid": [10, 10]},
        ),
        (
            {"grid": "O32", "area": [87.8638, 0, -87.8638, 357.6]},
            {"grid": [10, 10]},
        ),
        ({"grid": "N32", "global": 1}, {"grid": [10, 10]}),
        (
            {"grid": "N32", "global": 1, "area": [87.8638, 0, -87.8638, 357.188]},
            {"grid": [10, 10]},
        ),
        ({"grid": "H4"}, {"grid": [10, 10]}),
        ({"grid": "H4", "ordering": "ring"}, {"grid": [10, 10]}),
    ],
)
def test_local_gridspec_ok(gs_in, gs_out):
    DB = add_matrix_source(DB_PATH)
    r = DB.find_entry(gs_in, gs_out, "linear")
    assert r, f"gs_in={gs_in} gs_out={gs_out}"


@pytest.mark.parametrize(
    "gs_in,gs_out,err",
    [
        ({"grid": [1, 1]}, {"grid": [2, 2]}, None),
        ({"grid": [5, 5], "area": [90, 0, -90, 350]}, {"grid": [10, 10]}, None),
        ({"grid": [5, 5], "area": [90.001, 0, -90, 360]}, {"grid": [10, 10]}, None),
        ({"grid": [5, 5], "area": [90, 0, -89.0001, 360]}, {"grid": [10, 10]}, None),
        ({"grid": [5, 5], "area": [90, 0.001, -90, 360]}, {"grid": [10, 10]}, None),
        ({"grid": [5, 5], "area": [90, 0, -90, 359.999]}, {"grid": [10, 10]}, None),
        ({"grid": [5, 5], "area": [90, 10, -90, 370]}, {"grid": [10, 10]}, None),
        ({"grid": "G1280", "shape": 6599680}, {"grid": [10, 10]}, ValueError),
        ({"grid": "O32", "shape": 6599680}, {"grid": [10, 10]}, None),
        ({"grid": "O32", "area": [90, 0, -90, 359.999]}, {"grid": [10, 10]}, None),
        ({"grid": "O32", "area": [90, -0.1, -90, 360]}, {"grid": [10, 10]}, None),
        (
            {"grid": "O32", "area": [87.8638, 0, -87.8638, 357.7]},
            {"grid": [10, 10]},
            None,
        ),
        (
            {"grid": "O32", "area": [87.8638, 0.2, -87.8638, 357.5]},
            {"grid": [10, 10]},
            None,
        ),
        ({"grid": "N32", "shape": 6599680}, {"grid": [10, 10]}, None),
        ({"grid": "N32", "area": [90, 0, -90, 359.999]}, {"grid": [10, 10]}, None),
        ({"grid": "N32", "area": [90, -0.1, -90, 360]}, {"grid": [10, 10]}, None),
        ({"grid": "H4", "ordering": "any"}, {"grid": [10, 10]}, ValueError),
    ],
)
def test_local_gridspec_bad(gs_in, gs_out, err):
    DB = add_matrix_source(DB_PATH)
    if err:
        with pytest.raises(err):
            r = DB.find_entry(gs_in, gs_out, "linear")
    else:
        r = DB.find_entry(gs_in, gs_out, "linear")
        assert r is None, f"gs_in={gs_in} gs_out={gs_out}"



import argparse
import json
import logging
import os
import sys

here = os.path.dirname(__file__)
sys.path.insert(0, os.path.dirname(here))

from utils.matrix import download_index_file  # noqa

LOG = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO, format="%(levelname)s: %(message)s")

"""
This script merges multiple index.json files into a single index.json file.
"""


def to_file(d, target):
    with open(target, "w") as f:
        json.dump(d, f, indent=4)


def check_file(target, ref_num):
    with open(target) as f:
        db = json.load(f)
        assert len(db["matrix"]) == ref_num, f"{len(db['matrix'])} != {ref_num}"


def merge(input_dir, result_file, system_file=None, dry_run=False):
    db_res = {}
    total_num = 0
    if system_file is not None:
        with open(system_file) as f:
            db_res = json.load(f)
            total_num = len(db_res["matrix"])
        LOG.info(f"system index file: {system_file} number of entries={total_num}")

    LOG.info("load index file")
    # for f in os.scandir(input_dir):
    #     if f.is_dir() and f.name.startswith("matrices_"):
    #         index_file = os.path.join(f.path, "index.json")
    #         with open(index_file) as f:
    #             db = json.load(f)
    #             n = len(db["matrix"])
    #             total_num += n
    #             if not db_res:
    #                 db_res = db
    #             else:
    #                 db_res["matrix"].update(db["matrix"])
    #             LOG.info(f" loaded index file: {index_file} number of entries={n}")

    index_file = os.path.join(input_dir, "index.json")
    with open(index_file) as f:
        db = json.load(f)
        n = len(db["matrix"])
        total_num += n
        if not db_res:
            db_res = db
        else:
            db_res["matrix"].update(db["matrix"])
        LOG.info(f" loaded index file: {index_file} number of entries={n}")

    LOG.info(f"final index file: {result_file} number of entries={total_num}")
    if not dry_run:
        to_file(db_res, result_file)
        check_file(result_file, total_num)


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--mode",
        choices=["local", "system"],
        default="local",
        help="merge mode",
    )

    parser.add_argument(
        "--input-dir", default="", type=str, help="input root directory path"
    )
    parser.add_argument(
        "--output-dir", default="", type=str, help="output directory path"
    )

    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="perform test run without data processing",
    )

    # get args
    args = parser.parse_args()
    LOG.info(f"args={args}")

    if args.dry_run:
        LOG.warn("Running in dry-run mode!")

    if args.mode not in ["local", "system"]:
        raise ValueError(f"Invalid mode: {args.mode}")

    if not os.path.exists(args.input_dir):
        raise ValueError(f"input directory: {args.input_dir} doest not exist!")

    if args.output_dir == "":
        raise ValueError(f"Invalid output directory: {args.output_dir}")

    result_file = os.path.join(args.output_dir, "index.json")
    os.makedirs(os.path.dirname(result_file), exist_ok=True)

    if args.mode == "local":
        merge(args.input_dir, result_file, dry_run=args.dry_run)

    elif args.mode == "system":
        # fetch remote system index.json
        system_file = os.path.join(args.output_dir, "index.json.remote")
        download_index_file(system_file)
        merge(
            args.input_dir, result_file, system_file=system_file, dry_run=args.dry_run
        )


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        LOG.debug("Exception=KeyboardInterrupt")
        sys.exit(0)



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import hashlib
import json
from collections import defaultdict, namedtuple

from earthkit.regrid.db import SYS_DB as DB
from earthkit.regrid.gridspec import GridSpec

Specs = namedtuple("Specs", ["source", "target"])


BLOCK_COL_NUM = 3


def to_str(gs):
    if isinstance(gs, str):
        return gs
    return {"grid": gs["grid"]}


def make_gs_block(source, target):
    while len(target) % BLOCK_COL_NUM != 0:
        target.append("")

    step = len(target) // BLOCK_COL_NUM

    source_grid = source["grid"]

    CODE_LINE = "         {}"
    FIRST_COL = """
    * - .. code-block:: python

"""

    OTHER_COL = """
      - .. code-block:: python

"""

    txt = rf"""

{source_grid}
+++++++++++++++++++++++++

Source :ref:`gridspec <gridspec>`:

.. code-block:: python

    {to_str(source)}


Target :ref:`gridspec <gridspec>`\s available for source:

.. list-table::
    :header-rows: 0

"""

    for k in range(BLOCK_COL_NUM):
        txt += FIRST_COL if k == 0 else OTHER_COL
        for i in range(k * step, (k + 1) * step):
            txt += CODE_LINE.format(to_str(target[i])) + "\n"

    return txt


def match(gs, grid):
    for name, val in grid.items():
        if name == "type":
            continue
        if hasattr(gs, name):
            if getattr(gs, name) != val:
                return False
        else:
            return False
    return True


def build_gs_page(specs, grid, long_name):
    txt = f"""

.. include:: pre_gen_warn.rst

This page contains all the available target gridspecs for a given
**{long_name}** source :ref:`gridspec <gridspec>`.

"""

    # print(f"{grid=}\n")
    for _, v in specs.items():
        source = v.source
        target = v.target

        if source["type"] != grid["type"]:
            continue

        if match(source, grid):
            # print(f"->{source=}\n")
            # for t in target:
            # print(f"    {t=}\n")
            txt += make_gs_block(source, target)

    return txt


def load_matrix_index_file():
    specs = defaultdict(Specs)

    for _, entry in DB.index.items():
        gs_in = GridSpec.from_dict(entry["input"])
        gs_out = GridSpec.from_dict(entry["output"])

        if entry["interpolation"]["method"] == "linear":
            # key = dict(grid=gs_in["grid"])
            key = dict(gs_in)
            m = hashlib.sha256()
            m.update(json.dumps(key).encode("utf-8"))
            gs_id = m.hexdigest()

            if gs_id not in specs:
                specs[gs_id] = Specs(gs_in, [gs_out])
            else:
                specs[gs_id].target.append(gs_out)

    return specs


def execute(*args):
    specs = load_matrix_index_file()

    grid_type = args[0]

    gs = {}
    if grid_type == "reduced_gg_o":
        gs["type"] = "reduced_gg"
        gs["octahedral"] = True
    elif grid_type == "reduced_gg":
        gs["type"] = "reduced_gg"
        gs["octahedral"] = False
    elif grid_type == "healpix_ring":
        gs["type"] = "healpix"
        gs["ordering"] = "ring"
    elif grid_type == "healpix_nested":
        gs["type"] = "healpix"
        gs["ordering"] = "nested"
    else:
        gs["type"] = grid_type

    if len(args) >= 2:
        long_name = " ".join(args[1:])
        long_name = long_name.replace('"', "")
    else:
        long_name = grid_type

    # print(f"{gs=}\n")
    txt = build_gs_page(specs, gs, long_name)
    print(txt)


if __name__ == "__main__":
    execute()



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import argparse
import json
import logging
import os
import sys

import yaml

from earthkit.regrid.db import SYS_DB as DB

LOG = logging.getLogger(__name__)


"""Build a local matrix inventory by subsetting the default (remotely hosted) inventory."""


def subset_remote_db(
    conf_file, out_dir, strict=True, fail_on_missing=True, dry_run=False
):
    index_file = os.path.join(out_dir, "index.json")

    if not dry_run:
        os.makedirs(out_dir, exist_ok=True)
        if strict:
            if len(os.listdir(out_dir)) > 0:
                raise Exception("out_dir must be empty when strict=True")

    with open(conf_file) as f:
        items = yaml.safe_load(f)

    # res = dict(matrix={})
    # entries = []
    # inter = {}

    index = DB.subset_index(items, fail_on_missing=fail_on_missing)

    # copy matrices
    for _, entry in index.items():
        matrix_path = DB.copy_matrix_file(
            entry, out_dir, exist_ok=(not strict), dry_run=dry_run
        )

        LOG.info(f"  matrix_file: {os.path.relpath(matrix_path, out_dir)}")
        LOG.info("  matrix copied to out_dir")

    # save index file
    index_file = os.path.join(out_dir, "index.json")
    if not dry_run:
        with open(index_file, "w") as f:
            json.dump(index.to_raw(), f, indent=4)

    LOG.info(f"Index data saved to {index_file}")


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--conf", default="", type=str, help="configuration file path")
    parser.add_argument("--out-dir", default="", type=str, help="output directory path")

    parser.add_argument(
        "--allow-missing",
        action="store_true",
        help="do not fail when a required grid combination is not supported",
    )

    parser.add_argument(
        "--strict",
        action="store_true",
        help="fail when out_dir is not empty",
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="perform test run without data processing",
    )

    parser.add_argument(
        "--log-level",
        dest="log_level",
        choices=["debug", "info", "warn", "error", "critical"],
        default="info",
        help="set the logging level (default=info)",
    )

    # get args
    args = parser.parse_args()

    log_levels = {
        "debug": logging.DEBUG,
        "info": logging.INFO,
        "warn": logging.WARN,
        "error": logging.ERROR,
        "critical": logging.CRITICAL,
    }
    logging.basicConfig(
        level=log_levels[args.log_level], format="%(levelname)s: %(message)s"
    )

    # basic checks
    if not os.path.exists(args.conf):
        LOG.error(
            f"Config file path={args.conf} cannot be accessed! Use --conf to specify a valid path!"
        )
        sys.exit(1)

    if args.out_dir == "":
        LOG.error("No output directory specified! Use --out-dir to specify a path!")
        sys.exit(1)

    LOG.info(f"args={args}")

    subset_remote_db(
        args.conf,
        args.out_dir,
        strict=args.strict,
        fail_on_missing=not args.allow_missing,
        dry_run=args.dry_run,
    )

    LOG.info("DONE")


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        LOG.debug("Exception=KeyboardInterrupt")
        sys.exit(0)



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import hashlib
import json
import os

from earthkit.regrid.db import VERSION, MatrixIndex

from .mir import mir_cached_matrix_to_file


def regular_ll(entry):
    bb = entry["bbox"]
    d = {
        "grid": [entry["increments"][x] for x in ("west_east", "south_north")],
        "shape": [entry["nj"], entry["ni"]],
        "area": [bb["north"], bb["west"], bb["south"], bb["east"]],
    }
    if "global" in entry:
        d["global"] = entry["global"]
    return d


def reduced_gg(entry):
    pl = entry["pl"]
    G = "O" if pl[1] - pl[0] == 4 else "N"
    N = entry["N"]
    bb = entry["bbox"]

    d = {
        "grid": f"{G}{N}",
        "shape": [sum(pl)],
        "area": [bb["north"], bb["west"], bb["south"], bb["east"]],
    }

    if "global" in entry:
        d["global"] = entry["global"]

    return d


def healpix(entry):
    d = {"grid": entry["grid"], "ordering": entry["ordering"]}
    return d


def make_sha(d):
    m = hashlib.sha256()
    m.update(json.dumps(d, sort_keys=True).encode("utf-8"))
    return m.hexdigest()


def make_matrix(
    input_path, output_path, index_file=None, global_input=None, global_output=None
):
    with open(input_path) as f:
        entry = json.load(f)

    inter_ori = dict(entry["interpolation"])
    method = entry["interpolation"]["method"]
    uid = MatrixIndex.make_interpolation_uid(entry)
    if uid != method:
        entry["interpolation"]["_uid"] = uid

    # create output folder
    matrix_output_path = os.path.join(
        output_path,
        MatrixIndex.matrix_dir_name(entry),
        # f"{inter_engine}_{inter_version}_{inter_uid}"
    )
    os.makedirs(matrix_output_path, exist_ok=True)

    # create matrix
    cache_file = entry["matrix"].pop("cache_file")
    m = {}
    m["input"] = entry["input"]
    m["output"] = entry["output"]
    m["interpolation"] = inter_ori
    key = make_sha(m)
    name = key

    print(f"entry={entry}")
    npz_file = os.path.join(matrix_output_path, f"{name}.npz")
    mir_cached_matrix_to_file(cache_file, npz_file)

    if index_file is None:
        index_file = os.path.join(output_path, "index.json")

    if os.path.exists(index_file):
        with open(index_file) as f:
            index = json.load(f)
            if index.get("version", None) != VERSION:
                raise ValueError(f"{index_file=} version must be {VERSION}")
    else:
        index = {}
        index["version"] = VERSION
        index["matrix"] = {}

    def convert(x):
        proc = globals()[x["type"]]
        return proc(x)

    if global_input is not None and "global" not in entry["input"]:
        entry["input"]["global"] = 1 if global_input else 0

    if global_output is not None and "global" not in entry["output"]:
        entry["output"]["global"] = 1 if global_output else 0

    index["matrix"][key] = dict(
        input=convert(entry["input"]),
        output=convert(entry["output"]),
        interpolation=entry["interpolation"],
    )

    with open(index_file, "w") as f:
        json.dump(index, f, indent=4)

    print("Written", npz_file)
    print("Written", index_file)



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import earthkit.regrid


def test_version() -> None:
    v = earthkit.regrid.__version__
    print(f"earthkit-regrid version: {v}")
    assert v != "999"



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import ctypes
import datetime
import getpass
import hashlib
import json
import logging
import os
import platform
import shutil
import sqlite3
import tempfile
import threading
import time
from abc import ABCMeta, abstractmethod
from copy import deepcopy
from random import randrange

from filelock import FileLock

from earthkit.regrid.utils import humanize
from earthkit.regrid.utils.temporary import temp_directory

VERSION = 2
CACHE_DB = f"cache-{VERSION}.db"

LOG = logging.getLogger(__name__)


CONNECTION = None
CACHE = None

SETTINGS = {
    "cache-policy": "user",
    "user-cache-directory": os.path.join(
        tempfile.gettempdir(), f"earthkit-regrid-{getpass.getuser()}"
    ),
    "maximum-cache-size": 5 * 1024 * 1024 * 1024,
    "maximum-cache-disk-usage": 99,
    "url-download-timeout": 30,
    "check-out-of-date-urls": True,
    "download-out-of-date-urls": True,
    "temporary-cache-directory-root": None,
}


class DiskUsage:
    def __init__(self, path):
        path = os.path.realpath(path)
        self.path = path

        if platform.system() == "Windows":
            avail = ctypes.c_ulonglong()
            total = ctypes.c_ulonglong()
            free = ctypes.c_ulonglong()

            ctypes.windll.kernel32.GetDiskFreeSpaceExW(
                ctypes.c_wchar_p(path),
                ctypes.pointer(avail),
                ctypes.pointer(total),
                ctypes.pointer(free),
            )
            self.avail = avail.value
            self.total = total.value
            self.free = free.value
        else:
            st = os.statvfs(path)
            self.free = st.f_bfree * st.f_frsize
            self.total = st.f_blocks * st.f_frsize
            self.avail = st.f_bavail * st.f_frsize

        self.percent = int(
            float(self.total - self.avail) / float(self.total) * 100 + 0.5
        )

    def __repr__(self):
        return (
            f"DiskUsage(total={self.total},free={self.free},"
            f"avail={self.avail},percent={self.percent},path={self.path})"
        )


def disk_usage(path):
    return DiskUsage(path)


def default_serialiser(o):
    if isinstance(o, (datetime.date, datetime.datetime)):
        return o.isoformat()
    return json.JSONEncoder.default(o)


class Future:
    def __init__(self, func, args, kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self._condition = threading.Condition()
        self._ready = False
        self._result = None

    def execute(self):
        try:
            self._result = self.func(*self.args, **self.kwargs)
        except Exception as e:
            LOG.error(e)
            self._result = e
        with self._condition:
            self._ready = True
            self._condition.notify_all()

    def result(self):
        with self._condition:
            while not self._ready:
                self._condition.wait()
        if isinstance(self._result, Exception):
            raise self._result
        return self._result


class CacheManager(threading.Thread):
    def __init__(self):
        super().__init__(daemon=True)
        self._connection = None
        self._queue = []
        self._condition = threading.Condition()
        self._policy = EmptyCachePolicy()

    def run(self):
        while True:
            with self._condition:
                while len(self._queue) == 0:
                    self._condition.wait()
                s = self._queue.pop(0)
                self._condition.notify_all()
            s.execute()

    @property
    def connection(self):
        if self._policy.managed() and self._connection is None:
            self._connection = self.new_connection()

        return self._connection

    def new_connection(self):
        assert self._policy.managed()
        cache_dir = self._policy.directory()
        cache_db = os.path.join(cache_dir, CACHE_DB)
        LOG.debug("Cache database is %s", cache_db)
        connection = sqlite3.connect(cache_db)
        # So we can use rows as dictionaries
        connection.row_factory = sqlite3.Row

        # If you change the schema, change VERSION above
        connection.execute(
            """
            CREATE TABLE IF NOT EXISTS cache (
                    path          TEXT PRIMARY KEY,
                    owner         TEXT NOT NULL,
                    args          TEXT NOT NULL,
                    creation_date TEXT NOT NULL,
                    flags         INTEGER DEFAULT 0,
                    owner_data    TEXT,
                    last_access   TEXT NOT NULL,
                    type          TEXT,
                    parent        TEXT,
                    replaced      TEXT,
                    extra         TEXT,
                    expires       INTEGER,
                    accesses      INTEGER,
                    size          INTEGER);"""
        )
        return connection

    def enqueue(self, func, *args, **kwargs):
        with self._condition:
            s = Future(func, args, kwargs)
            self._queue.append(s)
            self._condition.notify_all()
            return s

    def _ensure_in_cache(self, path):
        assert self._policy.file_in_cache_directory(path), f"File not in cache {path}"

    def _settings_changed(self, policy):
        LOG.debug("Settings changed")
        self._policy = policy
        self._connection = None  # The user may have changed the cache directory
        self._check_cache_size()

    def _latest_date(self):
        """Returns the latest date to be used when purging the cache.
        So we do not purge files being downloaded.
        """
        with self.connection as db:
            latest = db.execute(
                "SELECT MIN(creation_date) FROM cache WHERE size IS NULL"
            ).fetchone()[0]
            if latest is None:
                latest = db.execute(
                    "SELECT MAX(creation_date) FROM cache WHERE size IS NOT NULL"
                ).fetchone()[0]
            if latest is None:
                latest = datetime.datetime.now()
            return latest

    def _purge_cache(self, matcher=None):
        if matcher is None:
            self._housekeeping(clean=True)
            # _update_cache(clean=True)
            self._decache(self._cache_size(), purge=True)
            return

        dump = self._dump_cache_database(matcher)
        for entry in dump:
            self._delete_entry(entry)

    def _cache_entries(self):
        result = []
        with self.connection as db:
            for n in db.execute("SELECT * FROM cache").fetchall():
                n = dict(n)
                n["args"] = json.loads(n["args"])
                try:
                    n["owner_data"] = json.loads(n["owner_data"])
                except Exception:
                    pass
                if os.path.exists(n["path"]):
                    result.append(n)
        return result

    def _update_entry(self, path, owner_data=None):
        self._ensure_in_cache(path)

        if os.path.isdir(path):
            kind = "directory"
            size = 0
            for root, _, files in os.walk(path):
                for f in files:
                    size += os.path.getsize(os.path.join(root, f))
        else:
            kind = "file"
            size = os.path.getsize(path)

        with self.connection as db:
            db.execute(
                "UPDATE cache SET size=?, type=?, owner_data=? WHERE path=?",
                (
                    size,
                    kind,
                    json.dumps(owner_data, default=default_serialiser),
                    path,
                ),
            )

    def _update_cache(self, clean=False):
        """Update cache size and size of each file in the database ."""
        with self.connection as db:
            update = []
            commit = False
            for n in db.execute("SELECT path FROM cache WHERE size IS NULL"):
                try:
                    path = n[0]
                    if os.path.isdir(path):
                        kind = "directory"
                        size = 0
                        for root, _, files in os.walk(path):
                            for f in files:
                                size += os.path.getsize(os.path.join(root, f))
                    else:
                        kind = "file"
                        size = os.path.getsize(path)
                    update.append((size, kind, path))
                except Exception:
                    if clean:
                        db.execute("DELETE from cache WHERE path=?", (path,))
                        commit = True

            if update:
                db.executemany("UPDATE cache SET size=?, type=? WHERE path=?", update)

            if update or commit:
                db.commit()

    def _housekeeping(self, clean=False):
        top = self._policy.directory()
        with self.connection as db:
            for name in os.listdir(top):
                if name == CACHE_DB:
                    continue

                full = os.path.join(top, name)
                count = db.execute(
                    "SELECT count(*) FROM cache WHERE path=?", (full,)
                ).fetchone()[0]

                if count > 0:
                    continue

                parent = None
                start = full.split(".")[0] + "%"
                for n in db.execute(
                    "SELECT path FROM cache WHERE parent IS NULL and path LIKE ?",
                    (start,),
                ).fetchall():
                    if full.startswith(n["path"]):
                        parent = n["path"]
                        break

                try:
                    s = os.stat(full)
                    if time.time() - s.st_mtime < 120:  # Two minutes
                        continue
                except OSError:
                    pass

                if parent is None:
                    LOG.warning(f"earthkit-data cache: orphan found: {full}")
                else:
                    LOG.debug(
                        f"earthkit-data cache: orphan found: {full} with parent {parent}"
                    )

                self._register_cache_file(
                    full,
                    "orphans",
                    None,
                    parent,
                )
        self._update_cache(clean=clean)

    def _delete_file(self, path):
        self._ensure_in_cache(path)

        try:
            if os.path.isdir(path) and not os.path.islink(path):
                shutil.rmtree(path)
            else:
                os.unlink(path)
        except Exception as e:
            print(e)
            LOG.exception("Deleting %s", path)

    def _entry_to_dict(self, entry):
        n = dict(entry)
        for k in ("args", "owner_data"):
            if k in n and isinstance(n[k], str):
                try:
                    n[k] = json.loads(n[k])
                except Exception:
                    LOG.debug("Cannot decode JSON %s", n[k])
                    pass
        return n

    def _delete_entry(self, entry):
        if isinstance(entry, str):
            entry = dict(
                path=entry,
                size=0,
                owner=None,
                args=None,
            )
            try:
                entry["size"] = os.path.getsize(entry["path"])
            except OSError:
                pass

        if entry["size"] is None:
            entry["size"] = 0

        path, size = (
            entry["path"],
            entry["size"],
        )

        # path, size, owner, args = (
        #     entry["path"],
        #     entry["size"],
        #     entry["owner"],
        #     entry["args"],
        # )

        # LOG.warning(
        #     "Deleting entry %s",
        #     json.dumps(
        #         self._entry_to_dict(entry), indent=4, default=default_serialiser
        #     ),
        # )
        total = 0

        # First, delete child files, e.g. unzipped data
        with self.connection as db:
            for child in db.execute("SELECT * FROM cache WHERE parent = ?", (path,)):
                total += self._delete_entry(child)

        if not os.path.exists(path):
            LOG.warning(f"cache file lost: {path}")
            with self.connection as db:
                db.execute("DELETE FROM cache WHERE path=?", (path,))
            return total

        # LOG.warning(f"earthkit-data cache: deleting {path} ({humanize.bytes(size)})")
        # LOG.warning(f"earthkit-data cache: {owner} {args}")
        self._delete_file(path)

        with self.connection as db:
            db.execute("DELETE FROM cache WHERE path=?", (path,))

        return total + size

    def _decache(self, bytes, purge=False):
        # _find_orphans()
        # _update_cache(clean=True)

        if bytes <= 0:
            return 0

        LOG.warning("earthkit-data cache: trying to free %s", humanize.bytes(bytes))

        total = 0

        with self.connection as db:
            latest = datetime.datetime.now() if purge else self._latest_date()

            for stmt in (
                "SELECT * FROM cache WHERE size IS NOT NULL AND owner='orphans' AND creation_date < ?",
                "SELECT * FROM cache WHERE size IS NOT NULL AND creation_date < ? ORDER BY last_access ASC",
            ):
                for entry in db.execute(stmt, (latest,)):
                    total += self._delete_entry(entry)
                    if total >= bytes:
                        LOG.warning(
                            "earthkit-data cache: freed %s from cache",
                            humanize.bytes(bytes),
                        )
                        return total

        LOG.warning("earthkit-data cache: could not free %s", humanize.bytes(bytes))

    def _register_cache_file(self, path, owner, args, parent=None):
        """Register a file in the cache

        Parameters
        ----------
        path : str
            Cache file to register
        owner : str
            Owner of the cache file (generally a source or a dataset)
        args : dict
            Dictionary to save with the file in the database, as json string.

        Returns
        -------
        changes :
            None or False if database does not need to be updated. TODO: clarify.
        """
        self._ensure_in_cache(path)

        with self.connection as db:
            now = datetime.datetime.now()

            args = json.dumps(args, default=default_serialiser)

            db.execute(
                """
                UPDATE cache
                SET accesses    = accesses + 1,
                    last_access = ?
                WHERE path=?""",
                (now, path),
            )

            changes = db.execute("SELECT changes()").fetchone()[0]

            if not changes:
                db.execute(
                    """
                    INSERT INTO cache(
                                    path,
                                    owner,
                                    args,
                                    creation_date,
                                    last_access,
                                    accesses,
                                    parent)
                    VALUES(?,?,?,?,?,?,?)""",
                    (path, owner, args, now, now, 1, parent),
                )

            return dict(
                db.execute("SELECT * FROM cache WHERE path=?", (path,)).fetchone()
            )

    def _cache_size(self):
        LOG.debug("cache_size")
        with self.connection as db:
            size = db.execute("SELECT SUM(size) FROM cache").fetchone()[0]
            if size is None:
                size = 0
            return size

    def _decache_file(self, path):
        self._delete_entry(path)

    def _check_cache_size(self):
        if self._policy.managed():
            # Check absolute limit
            maximum = self._policy.maximum_cache_size()
            if maximum is not None:
                size = self._cache_size()
                if size > maximum:
                    self._housekeeping()
                    self._decache(size - maximum)

            # Check relative limit
            usage = self._policy.maximum_cache_disk_usage()
            if usage is not None:
                size = self._cache_size()
                df = disk_usage(self._policy.directory())
                if df.percent > usage:
                    LOG.debug("Cache disk usage %s, limit %s", df.percent, usage)
                    self._housekeeping()
                    delta = (df.percent - usage) * df.total * 0.01
                    self._decache(delta)

    def _repr_html_(self):
        """Return a html representation of the cache .

        Returns
        -------
        str
            HTML status of the cache.
        """
        # html = [css("table")]
        html = []
        with self.new_connection() as db:
            for n in db.execute("SELECT * FROM cache"):
                n = dict(n)
                n["missing"] = not os.path.exists(n["path"])
                n["temporary"] = os.path.exists(n["path"] + ".tmp") or os.path.exists(
                    n["path"] + ".tmp.download"
                )  # TODO: decide how to handle temporary extension
                if n["size"] is None:
                    n["size"] = 0
                html.append("<table class='ek'>")
                html.append("<td><td colspan='2'>%s</td></tr>" % (n["path"],))

                for k in [x for x in n.keys() if x not in ("path", "owner_data")]:
                    v = humanize.bytes(n[k]) if k == "size" else n[k]
                    html.append("<td><td>%s</td><td>%s</td></tr>" % (k, v))
                html.append("</table>")
                html.append("<br>")
        return "".join(html)

    def _dump_cache_database(self, matcher=lambda x: True):
        result = []
        with self.connection as db:
            for d in db.execute("SELECT * FROM cache"):
                n = dict(d)
                for k in ("args", "owner_data"):
                    if n[k] is not None:
                        n[k] = json.loads(n[k])
                if matcher(n):
                    result.append(n)
        return result

    def _summary_dump_cache_database(self, matcher=lambda x: True):
        result = self._dump_cache_database(matcher=matcher)
        count = len(result)
        size = 0
        for r in result:
            size += r.get("size", 0)
        return count, size


class CachePolicy(metaclass=ABCMeta):
    CACHE_KEYS = [
        "cache-policy",
        "user-cache-directory",
        "temporary-cache-directory-root",
        "use-message-position-index-cache",
        "maximum-cache-disk-usage",
        "maximum-cache-size",
    ]

    OUTDATED_CHECK_KEYS = None
    _name = None

    def __init__(self):
        self._settings = {k: SETTINGS.get(k) for k in self.CACHE_KEYS}

    @property
    def name(self):
        return self._name

    @staticmethod
    def from_settings():
        name = SETTINGS.get("cache-policy")
        p = _cache_policies.get(name, None)
        if p is not None:
            return p()
        else:
            raise NotImplementedError(f"Unknown cache policy={name}")

    def outdated(self):
        return any(
            self._settings.get(k) != SETTINGS.get(k) for k in self.OUTDATED_CHECK_KEYS
        )

    def update(self):
        changed = False
        for k in self.CACHE_KEYS:
            if self._settings.get(k) != SETTINGS.get(k):
                changed = True
            if k not in self.OUTDATED_CHECK_KEYS:
                self._settings[k] = SETTINGS.get(k)
        return changed

    @abstractmethod
    def managed(self):
        pass

    @abstractmethod
    def directory(self):
        pass

    @abstractmethod
    def use_message_position_index_cache(self):
        pass

    @abstractmethod
    def is_cache_size_managed(self):
        pass

    @abstractmethod
    def maximum_cache_size(self):
        pass

    @abstractmethod
    def maximum_cache_disk_usage(self):
        pass

    def file_in_cache_directory(self, path):
        return path.startswith(self.directory())


class EmptyCachePolicy(CachePolicy):
    _name = "empty"

    def managed(self):
        return False

    def directory(self):
        return None

    def use_message_position_index_cache(self):
        return False

    def is_cache_size_managed(self):
        return False

    def maximum_cache_size(self):
        return None

    def maximum_cache_disk_usage(self):
        return None

    def __repr__(self):
        return self.__class__.__name__


class NoCachePolicy(CachePolicy):
    OUTDATED_CHECK_KEYS = ["cache-policy", "temporary-directory-root"]
    _name = "off"
    _dir = None

    def managed(self):
        return False

    def directory(self):
        if self._dir is None:
            if self._dir is None:
                root_dir = self._settings.get("temporary-directory-root")
                self._dir = temp_directory(dir=root_dir)
        return self._dir.path

    def use_message_position_index_cache(self):
        return False

    def is_cache_size_managed(self):
        return False

    def maximum_cache_size(self):
        return None

    def maximum_cache_disk_usage(self):
        return None

    def __repr__(self):
        return self.__class__.__name__


class UserCachePolicy(CachePolicy):
    OUTDATED_CHECK_KEYS = ["cache-policy", "user-cache-directory"]
    _name = "user"

    def __init__(self):
        super().__init__()
        path = self._settings.get("user-cache-directory")
        if not os.path.exists(path):
            os.makedirs(path, exist_ok=True)

    def managed(self):
        return True

    def directory(self):
        return self._settings.get("user-cache-directory")

    def use_message_position_index_cache(self):
        return self._settings.get("use-message-position-index-cache")

    def is_cache_size_managed(self):
        return (
            self.maximum_cache_size() is not None
            or self.maximum_cache_disk_usage() is not None
        )

    def maximum_cache_size(self):
        return self._settings.get("maximum-cache-size")

    def maximum_cache_disk_usage(self):
        return self._settings.get("maximum-cache-disk-usage")

    def __repr__(self):
        r = (
            f"{self.__class__.__name__}["
            f"user-cache-directory={self.directory()}"
            f", maximum-cache-size={self.maximum_cache_size()}"
            f", maximum-cache-disk-usage={self.maximum_cache_disk_usage()}"
            "]"
        )
        return r


class TmpCachePolicy(UserCachePolicy):
    OUTDATED_CHECK_KEYS = ["cache-policy", "temporary-cache-directory-root"]
    _name = "temporary"

    def __init__(self):
        super().__init__()
        root_dir = self._settings.get("temporary-cache-directory-root")
        self._dir = temp_directory(dir=root_dir)

    def directory(self):
        return self._dir.path


_cache_policies = {
    "off": NoCachePolicy,
    "temporary": TmpCachePolicy,
    "user": UserCachePolicy,
}


class Cache:
    """Class controlling the cache.

    See :ref:`caching` for details.
    """

    _created = False

    def __init__(self):
        if Cache._created:
            raise RuntimeError("Cannot create multiple instances of Cache")
        self._manager = None
        self._policy = None
        self._policy_lock = threading.Lock()
        self._manager_lock = threading.Lock()
        Cache._created = True

    @property
    def policy(self):
        r"""CachePolicy: Get the current cache policy."""
        with self._policy_lock:
            if self._policy is None:
                self._make_policy()
            return self._policy

    def _make_policy(self):
        if self._policy is None:
            self._policy = CachePolicy.from_settings()
            LOG.debug(f"Cache: created cache policy={self._policy}")
            if self._policy.managed():
                with self._manager_lock:
                    if self._manager is None:
                        self._manager = CacheManager()
                        self._manager.start()
                    self._call_manager_settings_changed()

    def _settings_changed(self):
        LOG.debug(
            "Cache: settings_changed, cache-policy=" + SETTINGS.get("cache-policy")
        )
        if self.policy.outdated():
            with self._policy_lock:
                # Check again, another thread/process may have modified the policy
                if self._policy.outdated():
                    self._policy = None
                    self._make_policy()
        elif self.policy.update() and self.policy.managed():
            with self._manager_lock:
                self._call_manager_settings_changed()

    def _call_manager(self, forget, name, *args, **kwargs):
        if self.policy.managed() and self._manager is not None:
            with self._manager_lock:
                name = "_" + name
                func = getattr(self._manager, name)
                if forget:
                    self._manager.enqueue(func, *args, **kwargs)
                    return None
                else:
                    s = self._manager.enqueue(func, *args, **kwargs)
                    return s.result()

    def _call_manager_settings_changed(self):
        s = self._manager.enqueue(
            self._manager._settings_changed, deepcopy(self._policy)
        )
        return s.result()

    def _dump_database(self, *args, **kwargs):
        return self._call_manager(False, "dump_cache_database", *args, **kwargs)

    def summary_dump_database(self, *args, **kwargs):
        """Return the number of items and total size of the cache.

        Does not work when the ``cache-policy`` is "off".

        Returns
        -------
        num: int
            number of items in the cache
        size: int
            total number of bytes stored in the cache


        >>> from earthkit.data import cache
        >>> cache.summary_dump_database()
        (40, 846785699)
        """
        return self._call_manager(False, "summary_dump_cache_database", *args, **kwargs)

    def _register_cache_file(self, *args, **kwargs):
        return self._call_manager(False, "register_cache_file", *args, **kwargs)

    def _update_entry(self, *args, **kwargs):
        return self._call_manager(False, "update_entry", *args, **kwargs)

    def _decache_file(self, *args, **kwargs):
        return self._call_manager(False, "decache_file", *args, **kwargs)

    def check_size(self, *args, **kwargs):
        """Check the cache size and trim it down when needed.

        Automatically runs when a new entry is added to the cache or the
        :ref:`cache_settings` change. Does not work when the
        ``cache-policy`` is "off".

        The algorithm includes three steps:

        - first, the cache size is determined
        - next, if the size is larger than the limit defined by
          the ``maximum-cache-size`` settings the oldest cache entries are
          removed until the desired size reached
        - finally, if the size is larger than the limit defined by the
          ``maximum-cache-disk-usage`` settings the oldest cache entries are
          removed until the desired size reached

        """
        return self._call_manager(True, "check_cache_size", *args, **kwargs)

    def size(self, *args, **kwargs):
        """Return the total number of bytes stored in the cache.

        Does not work when the ``cache-policy`` is "off".
        """
        return self._call_manager(False, "cache_size", *args, **kwargs)

    def entries(self, *args, **kwargs):
        """Dump the entries stored in the cache.

        Does not work when the ``cache-policy`` is "off".

        Returns
        -------
        list of dict:
            One dict per cache entry.
        """
        return self._call_manager(False, "cache_entries", *args, **kwargs)

    def purge(self, *args, **kwargs):
        """Delete entries from the cache.

        Does not work when the ``cache-policy`` is "off".

        Parameters
        ----------
        **kwargs: dict, optional
            Other keyword arguments:

            * matcher: callable
                Method to match the cache entries to delete. Its only argument
                is a cache entry and should return True if the entry is to be
                deleted.

        Examples
        --------
        Delete all entries.

        >>> from earthkit.data import cache
        >>> cache.purge()

        Delete all entries where the "owner" is "test_cache".

        >>> from earthkit.data import cache
        >>> cache.purge(matcher=lambda e: ["owner"] == "test_cache")

        """
        return self._call_manager(False, "purge_cache", *args, **kwargs)

    def _housekeeping(self, *args, **kwargs):
        return self._call_manager(False, "housekeeping", *args, **kwargs)

    def directory(self):
        """Return the path to the current (cache) directory.

        Returns
        -------
        str:
            The cache directory when ``cache-policy`` is "user" or "temporary".
            The temporary directory when ``cache-policy`` is "off".
        """
        return self.policy.directory()


CACHE = Cache()


def cache_file(
    owner: str,
    create,
    args,
    hash_extra=None,
    extension: str = ".cache",
    force=None,
    replace=None,
):
    """Creates a cache or temporary file in the folder specified by the cache-policy.

    Parameters
    ----------
    owner: str
        The owner of the cache file is generally the name of the source that
        called :func:`cache_file`.
    create: callable
        The method to create the contents of the cache file.
    args: list-like
        The parameters used to generate the cache key, which is also encoded into the
        file name and stored in the cache entry.
    extension: str
        Extension filename (such as ".nc" for NetCDF, etc.), by default ".cache".
    force: callable, bool
        Method or flag to decide whether an already existing cache file should
        be regenerated.

    Returns
    -------
    path: str
        Full path to the cache file.

    The behaviour depends on the cache policy:

    - If the cache-policy is ``user`` or ``temporary`` the file is created in the
    cache-directory and the relevant entries are added to the cache database
    using :py:meth:`Cache._register_cache_file`.

    - If the cache-policy is ``off`` the file is created in the temporary directory.
    No cache database and monitoring is available. The cache directory is merely
    serving as a temporary space.

    """
    managed = CACHE.policy.managed() and CACHE.directory() is not None

    if managed:
        m = hashlib.sha256()
        m.update(owner.encode("utf-8"))

        m.update(
            json.dumps(args, sort_keys=True, default=default_serialiser).encode("utf-8")
        )
        m.update(json.dumps(hash_extra, sort_keys=True).encode("utf-8"))
        m.update(json.dumps(extension, sort_keys=True).encode("utf-8"))

        if replace is not None:
            # Don't replace files that are not in the cache
            if not CACHE.policy.file_in_cache_directory(replace):
                replace = None

        path = os.path.join(
            CACHE.directory(),
            "{}-{}{}".format(
                owner.lower(),
                m.hexdigest(),
                extension,
            ),
        )

        record = CACHE._register_cache_file(path, owner, args)
        if os.path.exists(path):
            if callable(force):
                owner_data = record["owner_data"]
                if owner_data is not None:
                    owner_data = json.loads(owner_data)
                force = force(args, path, owner_data)

            if force:
                LOG.info(f"decache file by force: {path=}")
                CACHE._decache_file(path)
                record = CACHE._register_cache_file(path, owner, args)

        if not os.path.exists(path):
            lock = path + ".lock"

            with FileLock(lock):
                if not os.path.exists(
                    path
                ):  # Check again, another thread/process may have created the file
                    owner_data = create(path + ".tmp", args)
                    os.rename(path + ".tmp", path)
                    LOG.info(f"cache file created: {path=}")
                    CACHE._update_entry(path, owner_data)
                    CACHE.check_size()

            try:
                os.unlink(lock)
            except OSError:
                pass

    else:
        # path can be a file or a directory. We have to make the name unique.
        m = hashlib.sha256()
        m.update(datetime.datetime.now().isoformat().encode("utf-8"))
        m.update(str(randrange(10000000)).encode("utf-8"))

        path = os.path.join(
            CACHE.directory(),
            "{}-{}{}".format(
                owner.lower(),
                m.hexdigest(),
                extension,
            ),
        )

        if not os.path.exists(path):
            lock = path + ".lock"
            with FileLock(lock):
                if not os.path.exists(
                    path
                ):  # Check again, another thread/process may have created the file
                    owner_data = create(path + ".tmp", args)
                    os.rename(path + ".tmp", path)
            try:
                os.unlink(lock)
            except OSError:
                pass

    return path



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from earthkit.regrid.db import find


def interpolate(values, in_grid=None, out_grid=None, method="linear", **kwargs):
    interpolator = _find_interpolator(values)
    if interpolator is None:
        raise ValueError(f"Cannot interpolate data with type={type(values)}")

    return interpolator(
        values, in_grid=in_grid, out_grid=out_grid, method=method, **kwargs
    )


def _find_interpolator(values):
    for interpolator in INTERPOLATORS:
        if interpolator.match(values):
            return interpolator
    return None


def _interpolate(values, in_grid, out_grid, method, **kwargs):
    z, shape = find(in_grid, out_grid, method, **kwargs)

    if z is None:
        raise ValueError(f"No matrix found! {in_grid=} {out_grid=} {method=}")

    # This should check for 1D (GG) and 2D (LL) matrices
    values = values.reshape(-1, 1)

    values = z @ values

    return values.reshape(shape)


class NumpyInterpolator:
    @staticmethod
    def match(values):
        import numpy as np

        return isinstance(values, np.ndarray)

    def __call__(self, values, **kwargs):
        in_grid = kwargs.pop("in_grid")
        out_grid = kwargs.pop("out_grid")
        method = kwargs.pop("method")
        return _interpolate(values, in_grid, out_grid, method, **kwargs)


class FieldListInterpolator:
    @staticmethod
    def match(values):
        try:
            import earthkit.data

            return isinstance(values, earthkit.data.FieldList)
        except ImportError:
            return False

    def __call__(self, values, **kwargs):
        import earthkit.data

        ds = values
        in_grid = kwargs.pop("in_grid")
        if in_grid is not None:
            raise ValueError("in_grid cannot be used for FieldList interpolation")
        out_grid = kwargs.pop("out_grid")
        method = kwargs.pop("method")

        r = earthkit.data.FieldList()
        for f in ds:
            vv = f.to_numpy(flatten=True)
            v_res = _interpolate(
                vv,
                f.metadata().gridspec,
                out_grid,
                method,
                **kwargs,
            )
            md_res = f.metadata().override(gridspec=out_grid)
            r += ds.from_numpy(v_res, md_res)

        return r


INTERPOLATORS = [NumpyInterpolator(), FieldListInterpolator()]



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os
import sys

here = os.path.dirname(__file__)
sys.path.insert(0, os.path.dirname(here))

from utils.matrix import make_matrix  # noqa

logging.basicConfig(level=logging.DEBUG)

grids = [
    # ["5/5", "10/10"],
    # ["N32", "10/10"],
    # ["O32", "10/10"],
    # ["H4_ring", "10/10"],
    ["H4_nested", "10/10"],
]


methods = ["linear", "nearest-neighbour"]
build_dir = "_test_local_data_20240325"

index_file = os.path.join(build_dir, "index.json")

for method in methods:
    for g_in, g_out in grids:
        make_matrix(
            g_in,
            g_out,
            method,
            build_dir,
            index_file=index_file,
            download_index=False,
            delete_tmp_json=True,
        )
